import time

import requests


class RateLimitedSession(requests.Session):
    """
    A requests.Session that is rate limited.

    :param max_calls: Maximum number of calls per period
    :param period: Time period in seconds

    Example:

        session = RateLimitedSession(max_calls=10, period=1)
        session.get("https://example.com") # Will be rate limited
    """

    max_calls: int
    period: float

    _calls: int
    _reset_time: float

    def __init__(self, max_calls, period=1.0):
        self.max_calls = max_calls
        self.period = period

        self._calls = 0
        self._reset_time = time.time()
        super().__init__()

    def request(self, *args, **kwargs):
        if time.time() - self._reset_time > self.period:
            self._calls = 0
            self._reset_time = time.time()

        if self._calls < self.max_calls:
            self._calls += 1
        else:
            time.sleep(self.period - (time.time() - self._reset_time))
            # Note that this is self.request, not super().request.
            return self.request(*args, **kwargs)

        return super().request(*args, **kwargs)



# cspell:ignore jsons

from __future__ import annotations

import re
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Mapping, Optional, Union

from ..allspice import AllSpice
from ..apiobject import Ref, Repository
from .list_components import (
    ComponentAttributes,
    SupportedTool,
    infer_project_tool,
    list_components,
)

QUANTITY_COLUMN_NAME = "Quantity"


@dataclass
class ColumnConfig:
    """
    Configuration for a single column in the BOM.
    """

    class SortOrder(Enum):
        """
        Sort order for values in a column.
        """

        ASC = "asc"
        DESC = "desc"

    attributes: list[str] | str
    """
    The attributes of a part to use as the value for this column. If a list,
    each attribute in the list is tried in the order given, and the first one
    found is used. If a string, that attribute is used. If none of the given
    attributes are found, the column will be empty.
    """

    sort: Optional[SortOrder] = None
    """
    Sort order for the values in this column. If multiple columns have sort
    orders, the values will be sorted by the order of the columns in the
    columns mapping.

    The default is no sorting, and the order of the values in the BOM is not
    guaranteed.
    """

    remove_rows_matching: Optional[str] = None
    """
    A regex pattern to match against the values in this column. If a value
    matches this pattern, the **entire** row will be removed from the BOM.
    Filtering is performed after grouping.
    """

    grouped_values_sort: Optional[SortOrder] = None
    """
    If there is a grouping set, and this column is NOT part of the grouping,
    this will determine the sort order of the values in this column. For
    example, if the grouping is by "Reference" and "Value", and the column is
    "Designator", then the designators will be sorted in this order.

    The default is no sorting, and the order of the values in the group is not
    guaranteed.
    """

    grouped_values_separator: str = ", "
    """
    If there is a grouping set, and this column is NOT part of the grouping,
    this will determine the separator between the values in this column. For
    example, if the grouping is by "Reference" and "Value", and the column is
    "Designator", then the designators will be separated by this string.
    """

    grouped_values_allow_duplicates: bool = False
    """
    If there is a grouping set, and this column is NOT part of the grouping,
    this will determine if the values in this column should be deduplicated.
    For example, if the values "C1" and "C1" are found in the "Designator"
    column, and this is set to True, then only one "C1" will be included in
    the BOM. Otherwise, both will be included.
    """

    skip_in_output: bool = False
    """
    If True, this column will be skipped in the output BOM. This can be useful
    for columns that are used for grouping, sorting or filtering, but should
    not be included in the final BOM.
    """


ColumnsMapping = Mapping[str, ColumnConfig | list[str] | str]
"""
Configuration for the columns in the BOM. See `ColumnConfig` for a detailed
description of the configuration for each column. The keys in this dictionary
are the names of the columns in the BOM. The `str` and `list[str]` cases are
shorthand for `ColumnConfig` with the `attributes` field set to the given
value(s) and the other fields set to their defaults.
"""

BomEntry = dict[str, str]
Bom = list[BomEntry]


def generate_bom(
    allspice_client: AllSpice,
    repository: Repository,
    source_file: str,
    columns: ColumnsMapping,
    group_by: Optional[list[str]] = None,
    variant: Optional[str] = None,
    ref: Ref = "main",
    remove_non_bom_components: bool = True,
    design_reuse_repos: list[Repository] = [],
) -> Bom:
    """
    Generate a BOM for a project.

    :param allspice_client: The AllSpice client to use.
    :param repository: The repository to generate the BOM for.
    :param source_file: The path to the source file from the root of the
        repository. The source file must be a PrjPcb file for Altium projects a
        DSN file for OrCAD projects, or an SDAX file for System Capture
        projects. For example, if the source file is in the root of the
        repository and is named "Archimajor.PrjPcb", the path would be
        "Archimajor.PrjPcb"; if the source file is in a folder called
        "Schematics" and is named "Beagleplay.dsn", the path would be
        "Schematics/Beagleplay.dsn".
    :param columns: A mapping of the columns in the BOM to the attributes in the
        project. See `ColumnMapping` and `ColumnConfig` for a detailed
        description of the column configuration.

        Note that special attributes are added by this function depending on
        the project tool. For Altium projects, these are "_part_id",
        "_description", "_unique_id" and "_kind", which are the Library
        Reference, Description, Unique ID and Component Type respectively. For
        OrCAD and System Capture projects, "_name" is added, which is the name
        of the component, and "_reference" and "_logical_reference" may be
        added, which are the name of the component, and the logical reference
        of a multi-part component respectively.
    :param group_by: A list of columns to group the BOM by. If this is provided,
        the BOM will be grouped by the values of these columns.
    :param variant: The variant of the project to generate the BOM for. If this
        is provided, the BOM will be generated for the specified variant. If
        this is not provided, or is None, the BOM will be generated for the
        default variant. Variants are not supported for OrCAD projects.
    :param ref: The ref, i.e. branch, commit or git ref from which to take the
        project files. Defaults to "main".
    :param remove_non_bom_components: If True, components of types that should
        not be included in the BOM will be removed. Defaults to True. Only
        applicable for Altium projects.
    :return: A list of BOM entries. Each entry is a dictionary where the key is
        a column name and the value is the value for that column.
    """

    allspice_client.logger.info(
        f"Generating BOM for {repository.get_full_name()=} on {ref=} using {columns=}"
    )

    if group_by is not None:
        for group_column in group_by:
            if group_column not in columns:
                raise ValueError(f"Group by column {group_column} not found in selected columns")

    components = list_components(
        allspice_client,
        repository,
        source_file,
        variant,
        ref,
        combine_multi_part=True,
        design_reuse_repos=design_reuse_repos,
    )

    if remove_non_bom_components:
        project_tool = infer_project_tool(source_file)
        if project_tool == SupportedTool.ALTIUM:
            components = _remove_altium_non_bom_components(components)

    columns_mapping = {
        column_name: (
            column_config
            if isinstance(column_config, ColumnConfig)
            else ColumnConfig(attributes=column_config)
        )
        for column_name, column_config in columns.items()
    }

    mapped_components = _map_attributes(components, columns_mapping)
    bom = _group_entries(mapped_components, group_by, columns_mapping)
    bom = _filter_rows(bom, columns_mapping)
    bom = _sort_columns(bom, columns_mapping)
    bom = _skip_columns(bom, columns_mapping)

    return bom


def generate_bom_for_altium(
    allspice_client: AllSpice,
    repository: Repository,
    prjpcb_file: str,
    columns: ColumnsMapping,
    group_by: Optional[list[str]] = None,
    variant: Optional[str] = None,
    ref: Ref = "main",
    remove_non_bom_components: bool = True,
    design_reuse_repos: list[Repository] = [],
) -> Bom:
    """
    Generate a BOM for an Altium project.

    :param allspice_client: The AllSpice client to use.
    :param repository: The repository to generate the BOM for.
    :param prjpcb_file: The path to the PrjPcb project file from the root of the
        repository.
    :param columns: A mapping of the columns in the BOM to the attributes in the
        project. See `ColumnMapping` and `ColumnConfig` for a detailed
        description of the column configuration.

        Note that special attributes are added by this function, namely,
        "_part_id", "_description", "_unique_id" and "_kind", which are the
        Library Reference, Description, Unique ID and Component Type
        respectively.
    :param group_by: A list of columns to group the BOM by. If this is provided,
        the BOM will be grouped by the values of these columns.
    :param ref: The ref, i.e. branch, commit or git ref from which to take the
        project files. Defaults to "main".
    :param variant: The variant of the project to generate the BOM for. If this
        is provided, the BOM will be generated for the specified variant. If
        this is not provided, or is None, the BOM will be generated for the
        default variant.
    :param remove_non_bom_components: If True, components of types that should
        not be included in the BOM will be removed. Defaults to True.
    :return: A list of BOM entries. Each entry is a dictionary where the key is
        a column name and the value is the value for that column.
    """

    return generate_bom(
        allspice_client,
        repository,
        prjpcb_file,
        columns,
        group_by,
        variant,
        ref,
        remove_non_bom_components,
        design_reuse_repos=design_reuse_repos,
    )


def generate_bom_for_orcad(
    allspice_client: AllSpice,
    repository: Repository,
    dsn_path: str,
    columns: ColumnsMapping,
    group_by: Optional[list[str]] = None,
    ref: Ref = "main",
) -> Bom:
    """
    Generate a BOM for an OrCAD schematic.

    :param allspice_client: The AllSpice client to use.
    :param repository: The repository to generate the BOM for.
    :param dsn_path: The OrCAD DSN file. This can be a Content object returned
        by the AllSpice API, or a string containing the path to the file in the
        repo.
    :param columns: A mapping of the columns in the BOM to the attributes in the
        project. See `ColumnMapping` and `ColumnConfig` for a detailed
        description of the column configuration.

        Note that special attributes are added by this function, namely, "_name"
        is added, which is the name of the component.
    :param group_by: A list of columns to group the BOM by. If this is provided,
        the BOM will be grouped by the values of these columns.
    :param ref: The ref, i.e. branch, commit or git ref from which to take the
        project files. Defaults to "main".
    :return: A list of BOM entries. Each entry is a dictionary where the key is
        a column name and the value is the value for that column.
    """

    return generate_bom(
        allspice_client,
        repository,
        dsn_path,
        columns,
        group_by,
        ref=ref,
        remove_non_bom_components=False,
    )


def generate_bom_for_system_capture(
    allspice_client: AllSpice,
    repository: Repository,
    sdax_path: str,
    columns: ColumnsMapping,
    group_by: Optional[list[str]] = None,
    ref: Ref = "main",
) -> Bom:
    """
    Generate a BOM for a System Capture SDAX schematic.

    :param allspice_client: The AllSpice client to use.
    :param repository: The repository to generate the BOM for.
    :param sdax_path: The System Catpure SDAX schematic file. This can be a
        Content object returned by the AllSpice API, or a string containing the
        path to the file in the repo.
    :param columns: A mapping of the columns in the BOM to the attributes in the
        SDAX schematic. The attributes are tried in order, and the first one
        found is used as the value for that column.
    :param group_by: A list of columns to group the BOM by. If this is provided,
        the BOM will be grouped by the values of these columns.
    :param ref: The ref, i.e. branch, commit or git ref from which to take the
        project files. Defaults to "main".
    :return: A list of BOM entries. Each entry is a dictionary where the key is
        a column name and the value is the value for that column.
    """

    return generate_bom(
        allspice_client,
        repository,
        sdax_path,
        columns,
        group_by,
        ref=ref,
        remove_non_bom_components=False,
    )


def _get_first_matching_key_value(
    alternatives: Union[list[str], str],
    attributes: dict[str, str],
) -> Optional[str]:
    """
    Search for a series of alternative keys in a dictionary, and return the
    value of the first one found.
    """

    if isinstance(alternatives, str):
        alternatives = [alternatives]

    for alternative in alternatives:
        if alternative in attributes:
            return attributes[alternative]

    return None


def _map_attributes(
    components: list[ComponentAttributes],
    columns: dict[str, ColumnConfig],
) -> list[BomEntry]:
    """
    Map the attributes of the components to the columns of the BOM using the
    columns mapping. This takes a component as we get it from the JSON and
    returns a dict that can be used as a BOM entry.
    """

    return [
        {
            key: str(_get_first_matching_key_value(value.attributes, component) or "")
            for key, value in columns.items()
        }
        for component in components
    ]


def _group_entries(
    components: list[BomEntry],
    group_by: Optional[list[str]],
    columns_mapping: dict[str, ColumnConfig],
) -> list[BomEntry]:
    """
    Group components based on a list of columns. The order of the columns in the
    list will determine the order of the grouping.

    :returns: A list of rows which can be used as the BOM.
    """

    # If grouping is off, we just add a quantity of 1 to each component and
    # return early.
    if group_by is None or len(group_by) == 0:
        for component in components:
            component[QUANTITY_COLUMN_NAME] = "1"
        return components

    grouped_components = {}
    for component in components:
        key = tuple(component[column] for column in group_by)
        if key in grouped_components:
            grouped_components[key].append(component)
        else:
            grouped_components[key] = [component]

    rows = []

    for components in grouped_components.values():
        row = {}

        # We go through each column in the order they're defined, as that should
        # be the order in the BOM.
        for column, column_config in columns_mapping.items():
            # If we've grouped by this column, the value of the column is the
            # same for all components in the group, so we can just take the
            # value from the first component.
            if column in group_by:
                # The RHS here shouldn't fail as we've validated the group by
                # columns are all in the column selection.
                row[column] = components[0][column]
            # Otherwise, we need to combine the values from all components in
            # the group into one string.
            else:
                if column_config.grouped_values_allow_duplicates:
                    column_values = [str(component[column]) for component in components]
                else:
                    # dict.fromkeys retains the insertion order; set doesn't.
                    column_values = dict.fromkeys(
                        str(component[column]) for component in components
                    ).keys()
                sorted_values = _sort_values(column_values, column_config.grouped_values_sort)
                row[column] = column_config.grouped_values_separator.join(sorted_values)

        row[QUANTITY_COLUMN_NAME] = str(len(components))
        rows.append(row)

    return rows


def _remove_altium_non_bom_components(
    components: list[ComponentAttributes],
) -> list[ComponentAttributes]:
    """
    Filter out components of types that should not be included in the BOM.
    """

    return [
        component
        for component in components
        if component.get("_kind") not in {"NET_TIE_NO_BOM", "STANDARD_NO_BOM"}
    ]


def _sort_values(
    values: Iterable[str],
    sort_order: Optional[ColumnConfig.SortOrder],
) -> Iterable[str]:
    """
    Sort the values in a column based on the sort order.
    """

    if sort_order == ColumnConfig.SortOrder.ASC:
        return sorted(values)
    elif sort_order == ColumnConfig.SortOrder.DESC:
        return sorted(values, reverse=True)
    else:
        return values


def _sort_columns(bom_entries: Bom, columns_config: dict[str, ColumnConfig]) -> Bom:
    """
    Sort the BOM entries based on the sort order of the columns.
    """

    # It is possible (and perhaps faster) to sort the entries using `sorted`
    # and a custom `key` function. However, in multiple attempts to implement a
    # general solution that can handle a variable number of columns and sort
    # orders using `key`, I found it very difficult to read and understand what
    # it was doing. This solution should be slower in the pathological case
    # where *every* column has a sort order, but it is much easier to
    # understand and debug, and should be of comparable speed in the average
    # case where only a few columns have sort orders. Considering that the
    # number of columns is at most in the tens, even the pathological case
    # should be fine.

    # Get sortable columns in reverse order (least to most significant)
    sortable_columns = [
        (name, config) for name, config in reversed(columns_config.items()) if config.sort
    ]

    # Since we're using `sort` which mutates the list in place, we need to copy
    # the list to avoid modifying the original list.
    sorted_bom_entries = bom_entries.copy()

    for column_name, config in sortable_columns:
        reverse = config.sort == ColumnConfig.SortOrder.DESC
        sorted_bom_entries.sort(key=lambda entry: entry.get(column_name, ""), reverse=reverse)

    return sorted_bom_entries


def _filter_rows(bom_entries: Bom, columns_config: dict[str, ColumnConfig]) -> Bom:
    """
    Filter out rows based on the configuration of the columns.
    """

    columns_to_filter = {
        column: config.remove_rows_matching
        for column, config in columns_config.items()
        if config.remove_rows_matching
    }

    if len(columns_to_filter) == 0:
        return bom_entries

    return [
        row
        for row in bom_entries
        if not any(re.search(pattern, row[column]) for column, pattern in columns_to_filter.items())
    ]


def _skip_columns(bom_entries: Bom, columns_config: dict[str, ColumnConfig]) -> Bom:
    """
    Skip columns based on the configuration.
    """

    return [
        {
            column: value
            for column, value in row.items()
            if column not in columns_config or not columns_config[column].skip_in_output
        }
        for row in bom_entries
    ]



# cspell:ignore jsons

import configparser
import dataclasses
import functools
import pathlib
import re
import time
from dataclasses import dataclass
from enum import Enum
from logging import Logger
from typing import Optional

from ..allspice import AllSpice
from ..apiobject import Ref, Repository
from ..exceptions import NotYetGeneratedException

SLEEP_FOR_GENERATED_JSON = 1
"""The amount of time to sleep between attempts to fetch generated JSON files."""

MAX_RETRIES_FOR_GENERATED_JSON = 10
"""The maximum number of times to retry fetching generated JSON files."""

PCB_FOOTPRINT_ATTR_NAME = "PCB Footprint"

PART_REFERENCE_ATTR_NAME = "Part Reference"
"""The name of the part reference attribute in OrCAD projects."""

REPETITIONS_REGEX = re.compile(r"Repeat\((\w+),(\d+),(\d+)\)")

DESIGNATOR_COLUMN_NAME = "Designator"
"""The name of the designator attribute in Altium projects."""

LOGICAL_DESIGNATOR = "_logical_designator"

ComponentAttributes = dict[str, str]


class SupportedTool(Enum):
    """
    ECAD tools supported by list_components.
    """

    ALTIUM = "altium"
    ORCAD = "orcad"
    SYSTEM_CAPTURE = "system_capture"


class VariationKind(Enum):
    FITTED_MOD_PARAMS = 0
    NOT_FITTED = 1
    ALT_COMP = 2


@dataclass
class AltiumChildSheet:
    """
    A child sheet of a schematic in the hierarchy. The fields here are
    Altium-specific.
    """

    class ChildSheetKind(Enum):
        """
        The type of child sheet.
        """

        DOCUMENT = "Document"
        """
        A document in the project.
        """

        DEVICE_SHEET = "DeviceSheet"
        """
        A Device Sheet external to the project.
        """

    kind: ChildSheetKind
    """
    The type of child sheet referenced.
    """

    unique_id: str
    """
    The Unique ID of the SheetRef instance. Note that this may not be unique
    for child sheets, as the same sheetref can be expanded into multiple
    instances via channels.
    """

    name: str
    """
    The name of the child sheet. This can be a path for documents or a name for
    a device sheet.
    """

    channel_name: str
    """
    See https://www.altium.com/documentation/altium-designer/creating-multi-channel-design#the-repeat-keyword
    """


SchdocHierarchy = dict[str, list[AltiumChildSheet]]
"""
Mapping of the name of a sheet to its children.
"""


def list_components(
    allspice_client: AllSpice,
    repository: Repository,
    source_file: str,
    variant: Optional[str] = None,
    ref: Ref = "main",
    combine_multi_part: bool = False,
    design_reuse_repos: list[Repository] = [],
) -> list[ComponentAttributes]:
    """
    Get a list of all components in a schematic.


    Note that special attributes are added by this function depending on
    the project tool. For Altium projects, these are "_part_id",
    "_description", "_unique_id" and "_kind", which are the Library
    Reference, Description, Unique ID and Component Type respectively. For
    OrCAD and System Capture projects, "_name" is added, which is the name
    of the component, and "_reference" and "_logical_reference" may be
    added, which are the name of the component, and the logical reference
    of a multi-part component respectively.

    :param client: An AllSpice client instance.
    :param repository: The repository containing the schematic.
    :param source_file: The path to the schematic file from the repo root. The
        source file must be a PrjPcb file for Altium projects, a DSN file for
        OrCAD projects or an SDAX file for System Capture projects. For
        example, if the schematic is in the folder "Schematics" and the file is
        named "example.DSN", the path would be "Schematics/example.DSN".
    :param variant: The variant to apply to the components. If not None, the
        components will be filtered and modified according to the variant. Only
        applies to Altium projects.
    :param ref: Optional git ref to check. This can be a commit hash, branch
        name, or tag name. Default is "main", i.e. the main branch.
    :param combine_multi_part: If True, multi-part components will be combined
        into a single component.
    :return: A list of all components in the schematic. Each component is a
        dictionary with the keys being the attributes of the component and the
        values being the values of the attributes.
    """

    project_tool = infer_project_tool(source_file)
    match project_tool:
        case SupportedTool.ALTIUM:
            return list_components_for_altium(
                allspice_client,
                repository,
                source_file,
                variant=variant,
                ref=ref,
                combine_multi_part=combine_multi_part,
                design_reuse_repos=design_reuse_repos,
            )
        case SupportedTool.ORCAD:
            if variant:
                raise ValueError("Variant is not supported for OrCAD projects.")

            return list_components_for_orcad(
                allspice_client,
                repository,
                source_file,
                ref=ref,
                combine_multi_part=combine_multi_part,
            )
        case SupportedTool.SYSTEM_CAPTURE:
            if variant:
                raise ValueError("Variant is not supported for System Capture projects.")

            return list_components_for_system_capture(
                allspice_client,
                repository,
                source_file,
                ref=ref,
            )


def list_components_for_altium(
    allspice_client: AllSpice,
    repository: Repository,
    prjpcb_file: str,
    variant: Optional[str] = None,
    ref: Ref = "main",
    combine_multi_part: bool = False,
    design_reuse_repos: list[Repository] = [],
) -> list[ComponentAttributes]:
    """
    Get a list of all components in an Altium project.

    :param client: An AllSpice client instance.
    :param repository: The repository containing the Altium project.
    :param prjpcb_file: The path to the PrjPcb file from the repo root. For
        example, if the PrjPcb file is in the folder "Project" and the file
        is named "example.prjpcb", the path would be "Project/example.prjpcb".
    :param variant: The variant to apply to the components. If not None, the
        components will be filtered and modified according to the variant.
    :param ref: Optional git ref to check. This can be a commit hash, branch
        name, or tag name. Default is "main", i.e. the main branch.
    :param combine_multi_part: If True, multi-part components will be combined
        into a single component.
    :return: A list of all components in the Altium project. Each component is
        a dictionary with the keys being the attributes of the component and the
        values being the values of the attributes. Additionally, `_part_id`,
        `_description`, `_unique_id`, and `_kind` attributes are added to each
        component to store the Library Reference, Description, Unique ID, and
        Component Type respectively.
    """

    allspice_client.logger.info(f"Fetching {prjpcb_file=}")

    # Altium adds the Byte Order Mark to UTF-8 files, so we need to decode the
    # file content with utf-8-sig to remove it.
    prjpcb_file_contents = repository.get_raw_file(prjpcb_file, ref=ref).decode("utf-8-sig")

    prjpcb_ini = configparser.ConfigParser()
    prjpcb_ini.read_string(prjpcb_file_contents)

    if variant is not None:
        try:
            variant_details = _extract_variations(variant, prjpcb_ini)
        except ValueError:
            raise ValueError(
                f"Variant {variant} not found in PrjPcb file. "
                "Please check the name of the variant."
            )
    else:
        # Ensuring variant_details is always bound, even if it is not used.
        variant_details = None

    project_documents, device_sheets = _extract_schdoc_list_from_prjpcb(prjpcb_ini)
    allspice_client.logger.info("Found %d SchDoc files", len(project_documents))
    allspice_client.logger.info("Found %d Device Sheet files", len(device_sheets))

    if not project_documents:
        raise ValueError("No Project Documents found in the PrjPcb file.")

    try:
        annotations_data = _fetch_and_parse_annotation_file(repository, prjpcb_file, ref)
        allspice_client.logger.info("Found annotations file, %d entries", len(annotations_data))
    except Exception as e:
        if device_sheets:
            allspice_client.logger.warning("Failed to fetch annotations file: %s", e)
            allspice_client.logger.warning("Component designators may not be correct.")
        annotations_data = {}

    # Mapping of schdoc file paths from the project file to their JSON
    schdoc_jsons: dict[str, dict] = {}
    # Mapping of device sheet *names* to their JSON.
    device_sheet_jsons: dict[str, dict] = {}

    # When working with Device Sheets, there are three different paths to keep
    # in mind:
    #
    # 1. The path as given in the PrjPcb file: Used to find a possible match in
    #    the current repo, e.g. if it is a monorepo. If not found, we get the
    #    basename and use that to match in the design reuse repos, yielding (2)
    #    after which this is no longer used.
    # 2. The path of the file in the design reuse repository: Used for fetching
    #    the JSON, after which this is no longer used.
    # 3. The name stored in the `filename` property of the SheetRef: This is the
    #    stem of the path. We need to match a SheetRef to the device sheet it is
    #    referring to using this, which is why `device_sheet_jsons` uses these
    #    as the keys.

    for schdoc_file in project_documents:
        schdoc_path_from_repo_root = _resolve_prjpcb_relative_path(schdoc_file, prjpcb_file)

        schdoc_json = _fetch_generated_json(
            repository,
            schdoc_path_from_repo_root,
            ref,
        )
        schdoc_jsons[schdoc_path_from_repo_root] = schdoc_json

    for device_sheet in device_sheets:
        device_sheet_repo, device_sheet_path = _find_device_sheet(
            device_sheet,
            repository,
            prjpcb_file,
            design_reuse_repos,
            allspice_client.logger,
        )
        device_sheet_json = _fetch_generated_json(
            device_sheet_repo,
            device_sheet_path.as_posix(),
            # Note the default branch here - we can't assume the same ref is
            # available.
            device_sheet_repo.default_branch,
        )
        device_sheet_jsons[device_sheet_path.stem] = device_sheet_json

    independent_sheets, hierarchy = _build_schdoc_hierarchy(schdoc_jsons, device_sheet_jsons)

    unique_ids_mapping = _create_unique_ids_mapping(prjpcb_ini)
    if device_sheets:
        hierarchy = _correct_device_sheet_reference_unique_ids(
            hierarchy,
            prjpcb_ini,
            allspice_client.logger,
        )

    allspice_client.logger.debug("Independent sheets: %s", independent_sheets)
    allspice_client.logger.debug("Hierarchy: %s", hierarchy)

    # Now we can build a combined mapping of documents and device sheets:
    sheets_to_components = {}
    for schdoc_file, schdoc_json in schdoc_jsons.items():
        schdoc_components = [
            value
            for value in schdoc_json.values()
            if isinstance(value, dict) and value.get("type") == "Component"
        ]
        sheets_to_components[schdoc_file] = schdoc_components
    for device_sheet_name, device_sheet_json in device_sheet_jsons.items():
        device_sheet_components = [
            value
            for value in device_sheet_json.values()
            if isinstance(value, dict) and value.get("type") == "Component"
        ]
        sheets_to_components[device_sheet_name] = device_sheet_components

    components = []

    for independent_sheet in independent_sheets:
        components.extend(
            _extract_components_altium(
                independent_sheet,
                sheets_to_components,
                hierarchy,
                parent_sheet_id="",
                current_sheet=None,
            )
        )

    # At this stage, we have components where the designators are not final.

    if annotations_data:
        components = _apply_annotation_file(
            components,
            annotations_data,
            unique_ids_mapping,
            allspice_client.logger,
        )
    else:
        components = _compute_repetitions(components)

    if combine_multi_part:
        # Multi part components must be combined *after* we've processed
        # repetitions and before we apply variations. This is because each
        # repetition of a multi-part component is treated as a separate
        # component, and they can be present across sheets. We need to combine
        # them into a single component before applying variations, as Altium
        # variations will apply to the combined component.
        components = _combine_multi_part_components_for_altium(components)

    if variant is not None:
        if variant_details is None:
            # This should never happen, but mypy doesn't know that.
            raise ValueError(f"Variant {variant} not found in PrjPcb file.")

        components = _apply_variations(components, variant_details, allspice_client.logger)

    return _filter_blank_components(components, allspice_client.logger)


def list_components_for_orcad(
    allspice_client: AllSpice,
    repository: Repository,
    dsn_path: str,
    ref: Ref = "main",
    combine_multi_part: bool = False,
) -> list[ComponentAttributes]:
    """
    Get a list of all components in an OrCAD DSN schematic.

    :param client: An AllSpice client instance.
    :param repository: The repository containing the OrCAD schematic.
    :param dsn_path: The path to the OrCAD DSN file from the repo root. For
        example, if the schematic is in the folder "Schematics" and the file
        is named "example.dsn", the path would be "Schematics/example.dsn".
    :param ref: Optional git ref to check. This can be a commit hash, branch
        name, or tag name. Default is "main", i.e. the main branch.
    :param combine_multi_part: If True, multi-part components will be combined
        into a single component.
    :return: A list of all components in the OrCAD schematic. Each component is
        a dictionary with the keys being the attributes of the component and the
        values being the values of the attributes. A `_name` attribute is added
        to each component to store the name of the component.
    """

    components = _list_components_multi_page_schematic(allspice_client, repository, dsn_path, ref)

    if combine_multi_part:
        components = _combine_multi_part_components_for_orcad(components)

    return components


def list_components_for_system_capture(
    allspice_client: AllSpice,
    repository: Repository,
    sdax_path: str,
    ref: Ref = "main",
) -> list[ComponentAttributes]:
    """
    Get a list of all components in a System Capture SDAX schematic.

    :param client: An AllSpice client instance.
    :param repository: The repository containing the System Capture schematic.
    :param sdax_path: The path to the System Capture SDAX file from the repo
        root. For example, if the schematic is in the folder "Schematics" and
        the file is named "example.sdax", the path would be
        "Schematics/example.sdax".
    :param ref: Optional git ref to check. This can be a commit hash, branch
        name, or tag name. Default is "main", i.e. the main branch.
    """

    return _list_components_multi_page_schematic(allspice_client, repository, sdax_path, ref)


def infer_project_tool(source_file: str) -> SupportedTool:
    """
    Infer the ECAD tool used in a project from the file extension.
    """

    if source_file.lower().endswith(".prjpcb"):
        return SupportedTool.ALTIUM
    elif source_file.lower().endswith(".dsn"):
        return SupportedTool.ORCAD
    elif source_file.lower().endswith(".sdax"):
        return SupportedTool.SYSTEM_CAPTURE
    else:
        raise ValueError("""
The source file for generate_bom must be:

- A PrjPcb file for Altium projects; or
- A DSN file for OrCAD projects; or
- An SDAX file for System Capture projects.
        """)


def _list_components_multi_page_schematic(
    allspice_client: AllSpice,
    repository: Repository,
    schematic_path: str,
    ref: Ref,
) -> list[dict[str, str]]:
    """
    Internal function for getting all components from a multi-page schematic.

    This pattern is followed by OrCAD and System Capture, and potentially other
    formats in the future.
    """

    allspice_client.logger.debug(
        f"Listing components in {schematic_path=} from {repository.get_full_name()} on {ref=}"
    )

    # Get the generated JSON for the schematic.
    schematic_json = _fetch_generated_json(repository, schematic_path, ref)
    pages = schematic_json["pages"]
    components = []

    for page in pages:
        for component in page["components"].values():
            component_attributes = _component_attributes_multi_page(component)
            components.append(component_attributes)

    return _filter_blank_components(components, allspice_client.logger)


def _fetch_generated_json(repo: Repository, file_path: str, ref: Ref) -> dict:
    attempts = 0
    while attempts < MAX_RETRIES_FOR_GENERATED_JSON:
        try:
            return repo.get_generated_json(file_path, ref=ref)
        except NotYetGeneratedException:
            attempts += 1
            time.sleep(SLEEP_FOR_GENERATED_JSON)

    raise TimeoutError(
        f"Failed to fetch JSON for {file_path} after {MAX_RETRIES_FOR_GENERATED_JSON} attempts."
    )


@functools.cache
def _fetch_all_files_in_repo(repo: Repository) -> list[str]:
    """
    Fetch a list of all files in a repository at the default branch.

    :param repo: The repository.
    :returns: A list of all file paths in the repo.
    """

    tree = repo.get_tree(recursive=True)

    return [file.path for file in tree]


def _extract_schdoc_list_from_prjpcb(
    prjpcb_ini: configparser.ConfigParser,
) -> tuple[set[str], set[str]]:
    """
    Extract sets of all schematic files used in this project.

    :param prjpcb_ini: The contents of the PrjPcb file as a ConfigParser.
    :returns: two sets. The first set is of relative paths from the project
        file to all the Schematic Documents in this project, and the second is
        of relative paths to device sheets from the project file.
    """

    schdoc_files = set()
    device_sheets = set()

    for section in prjpcb_ini.sections():
        if section.casefold().startswith("document"):
            document_path = prjpcb_ini[section].get("DocumentPath")
            if document_path and document_path.casefold().endswith(".schdoc"):
                schdoc_files.add(document_path)
        elif section.casefold().startswith("devicesheet"):
            document_path = prjpcb_ini[section].get("DocumentPath")
            if document_path and document_path.casefold().endswith(".schdoc"):
                device_sheets.add(document_path)

    return (schdoc_files, device_sheets)


def _resolve_prjpcb_relative_path(schdoc_path: str, prjpcb_path: str) -> str:
    """
    Convert a relative path to the SchDoc file to an absolute path from the git
    root based on the path to the PrjPcb file.
    """

    # The paths in the PrjPcb file are Windows paths, and ASH will store the
    # paths as Posix paths. We need to resolve the SchDoc path relative to the
    # PrjPcb path (which is a Posix Path, since it is from ASH), and then
    # convert the result into a posix path as a string for use in ASH.
    schdoc = pathlib.PureWindowsPath(schdoc_path)
    prjpcb = pathlib.PurePosixPath(prjpcb_path)
    return (prjpcb.parent / schdoc).as_posix()


def _build_schdoc_hierarchy(
    schematic_document_jsons: dict[str, dict],
    device_sheet_jsons: dict[str, dict],
) -> tuple[set[str], SchdocHierarchy]:
    """
    Build a hierarchy of sheets from a mapping of sheet names to the references
    of their children.

    :param document_jsons: A mapping of document sheet paths from project root
        to their JSON.
    :param device_sheet_jsons: A mapping of device sheet *names* to their JSON.
    :returns: The output of this function is a tuple of two values:

    1. A set of "independent" sheets, which can be taken to be roots of the
       hierarchy.
    2. A mapping of each sheet that has children to a list of tuples, where
       each tuple is a child sheet and the number of repetitions of that child
       sheet in the parent sheet. If a sheet has no children and is not a child
       of any other sheet, it will be mapped to an empty list.
    """

    hierarchy: SchdocHierarchy = {}

    schematic_document_entries = {
        schdoc_file: [value for value in schdoc_json.values() if isinstance(value, dict)]
        for schdoc_file, schdoc_json in schematic_document_jsons.items()
    }
    schematic_document_refs = {
        schdoc_file: [entry for entry in entries if entry.get("type") == "SheetRef"]
        for schdoc_file, entries in schematic_document_entries.items()
    }

    device_sheet_entries = {
        device_sheet_name: [
            value for value in device_sheet_json.values() if isinstance(value, dict)
        ]
        for device_sheet_name, device_sheet_json in device_sheet_jsons.items()
    }
    device_sheet_refs = {
        device_sheet_name: [entry for entry in entries if entry.get("type") == "SheetRef"]
        for device_sheet_name, entries in device_sheet_entries.items()
    }

    # We start by assuming all the document sheets are independent. Device
    # sheets cannot be independent, as they must be referred to by a document
    # sheet or another device sheet.
    independent_sheets = set(schematic_document_refs.keys())
    # This is what we'll use to compare with the sheet names in repetitions.
    schematic_document_names_downcased = {sheet.casefold(): sheet for sheet in independent_sheets}
    device_sheet_names_downcased = {sheet.casefold(): sheet for sheet in device_sheet_refs.keys()}

    # First, we'll build the hierarchy for device sheets, since they cannot
    # point to document sheets.
    for device_sheet_name, device_sheet_refs in device_sheet_refs.items():
        if not device_sheet_refs:
            continue

        repetitions = _extract_repetitions(device_sheet_refs)
        for child_sheet in repetitions:
            # child_sheet should just be a filename
            child_name = device_sheet_names_downcased[child_sheet.name.casefold()]
            # We have to replace the name in the sheet ref with the name in the
            # project file we expect
            child_sheet = dataclasses.replace(
                child_sheet,
                name=child_name,
                kind=AltiumChildSheet.ChildSheetKind.DEVICE_SHEET,
            )
            hierarchy.setdefault(device_sheet_name, []).append(child_sheet)

    # Now for the document sheets, which can point to both other document
    # sheets and device sheets
    for schematic_document_sheet, refs in schematic_document_refs.items():
        if not refs or len(refs) == 0:
            continue

        repetitions = _extract_repetitions(refs)

        for child_sheet in repetitions:
            child_path = _resolve_child_relative_path(
                child_sheet.name, schematic_document_sheet
            ).casefold()
            if child_path in schematic_document_names_downcased:
                child_name = schematic_document_names_downcased[child_path]
            else:
                # Note the `child_sheet` below - we use the bare text without
                # any path resolution for device sheets.
                child_name = device_sheet_names_downcased[child_sheet.name.casefold()]
                # Now we know that this child sheet is a Device Sheet, we can
                # replace the kind.
                child_sheet = dataclasses.replace(
                    child_sheet,
                    kind=AltiumChildSheet.ChildSheetKind.DEVICE_SHEET,
                )

            # We have to replace the name in the sheet ref with the name in the project file we expect
            child_sheet = dataclasses.replace(child_sheet, name=child_name)
            hierarchy.setdefault(schematic_document_sheet, []).append(child_sheet)
            independent_sheets.discard(child_name)

    return (independent_sheets, hierarchy)


def _resolve_child_relative_path(child_path: str, parent_path: str) -> str:
    """
    Converts a relative path in a sheet ref to a POSIX relative path from the
    prjpcb file.

    The returned path is POSIX as we convert them to POSIX paths when fetching
    JSONs.
    """

    child = pathlib.PureWindowsPath(child_path)
    parent = pathlib.PureWindowsPath(parent_path)

    return (parent.parent / child).as_posix()


def _extract_repetitions(sheet_refs: list[dict]) -> list[AltiumChildSheet]:
    """
    Takes a list of sheet references and returns all child sheets, which include
    the repetition count for each child sheet.
    """

    repetitions: list[AltiumChildSheet] = []

    for sheet_ref in sheet_refs:
        sheet_name = (sheet_ref.get("sheet_name", {}) or {}).get("name", "") or ""
        try:
            sheet_file_name = sheet_ref["filename"]
        except Exception:
            raise ValueError(f"Could not find sheet filename in {sheet_ref=}")
        if sheet_file_name is None:
            raise ValueError(
                "Sheet filename is null in for a sheet. Please check sheet references in this "
                "project for an empty file path."
            )

        if match := REPETITIONS_REGEX.search(sheet_name):
            channel_identifier = match.group(1)
            start = int(match.group(2))
            end = int(match.group(3))

            for channel_index in range(start, end + 1):
                repetitions.append(
                    AltiumChildSheet(
                        # At this stage, we assume everything is a document, as
                        # we can't know if it is a device sheet or not.
                        kind=AltiumChildSheet.ChildSheetKind.DOCUMENT,
                        name=sheet_file_name,
                        # The UniqueID of this specific instance has to include
                        # the channel index before it.
                        unique_id=f"{channel_index}{sheet_ref['unique_id']}",
                        channel_name=f"{channel_identifier}{channel_index}",
                    )
                )
        else:
            repetitions.append(
                AltiumChildSheet(
                    # At this stage, we assume everything is a document, as
                    # we can't know if it is a device sheet or not.
                    kind=AltiumChildSheet.ChildSheetKind.DOCUMENT,
                    name=sheet_file_name,
                    unique_id=sheet_ref["unique_id"],
                    channel_name=sheet_name,
                )
            )

    return repetitions


def _component_attributes_altium(component: dict) -> ComponentAttributes:
    """
    Extract the attributes of a component into a dict.

    This also adds some top-level properties of the component that are not
    attributes into the dict.
    """

    attributes = {}

    for key, value in component["attributes"].items():
        attributes[key] = value["text"]

    # The designator attribute has a `value` key which contains the unchanged
    # designator from the schematic file. This is useful when combining multi-
    # part components.
    attributes[LOGICAL_DESIGNATOR] = component["attributes"][DESIGNATOR_COLUMN_NAME]["value"]

    if "part_id" in component:
        attributes["_part_id"] = component["part_id"]
    if "name" in component:
        attributes["_name"] = component["name"]
    if "description" in component:
        attributes["_description"] = component["description"]
    if "unique_id" in component:
        attributes["_unique_id"] = component["unique_id"]
    if "kind" in component:
        attributes["_kind"] = component["kind"]
    if "part_count" in component:
        attributes["_part_count"] = component["part_count"]
        attributes["_current_part_id"] = component["current_part_id"]
    if "pins" in component:
        attributes["_pins"] = component["pins"]

    return attributes


def _component_attributes_multi_page(component: dict) -> ComponentAttributes:
    """
    Extract attributes of components from a multi-page document into a dict.

    This also adds some of the top-level properties of the component that
    are not attributes into the dict.
    """

    component_attributes = {}

    component_attributes["_name"] = component["name"]
    if "reference" in component:
        component_attributes["_reference"] = component.get("reference")
    if "logical_reference" in component:
        component_attributes["_logical_reference"] = component.get("logical_reference")
    if "pins" in component:
        # Multi-sheet documents have pins as an object of ids to pin objects.
        # The id is also inside the pin object, so we can flatten this into an
        # array, which is also compatible with how Altium outputs them.
        component_attributes["_pins"] = list(component["pins"].values())

    for attribute in component["attributes"].values():
        component_attributes[attribute["name"]] = attribute["value"]

    return component_attributes


def _letters_for_repetition(rep: int) -> str:
    """
    Generate the letter suffix for a repetition number. If the repetition is
    more than 26, the suffix will be a combination of letters.
    """

    first = ord("A")
    suffix = ""

    while rep > 0:
        u = (rep - 1) % 26
        letter = chr(u + first)
        suffix = letter + suffix
        rep = (rep - u) // 26

    return suffix


def _extract_components_altium(
    sheet_name: str,
    sheets_to_entries: dict[str, list[dict]],
    hierarchy: SchdocHierarchy,
    parent_sheet_id: str,
    current_sheet: AltiumChildSheet | None = None,
) -> list[ComponentAttributes]:
    """
    Extract the components from a sheet in an Altium project.

    :param sheet_name: The name of the sheet to extract components from. This
        is required because independent sheets do not have an AltiumChildSheet
        instance, and we need the name to look up the components.
    :param sheets_to_entries: A mapping of sheet names to the entries in the
        JSON for that sheet.
    :param hierarchy: The hierarchy of the sheets in the project.
    :param parent_sheet_id: The Unique ID of the parent sheet in the hierarchy.
        For independent sheets, this should be empty.
    :param current_sheet: The AltiumChildSheet instance for the current sheet.
        If this is an independent sheet, this should be None.
    :returns: A list of components in the sheet. The designators of the
        components here are not final and should be patched, either with the
        Annotation file or through the repetitions logic.
    """

    components = []

    if current_sheet is not None:
        current_sheet_id = f"{parent_sheet_id}\\{current_sheet.unique_id}"
    else:
        current_sheet_id = parent_sheet_id

    if sheet_name not in sheets_to_entries:
        return components

    for entry in sheets_to_entries[sheet_name]:
        if entry["type"] != "Component":
            continue

        component = _component_attributes_altium(entry)
        component_unique_id = f"{current_sheet_id}\\{component['_unique_id']}"
        component["_unique_id"] = component_unique_id
        if current_sheet is not None:
            component["_channel_name"] = current_sheet.channel_name

        components.append(component)

    if sheet_name not in hierarchy:
        return components

    for child_sheet in hierarchy[sheet_name]:
        child_components = _extract_components_altium(
            child_sheet.name,
            sheets_to_entries,
            hierarchy,
            current_sheet_id,
            child_sheet,
        )

        components.extend(child_components)

    return components


def _combine_multi_part_components_for_altium(
    components: list[dict[str, str]],
) -> list[dict[str, str]]:
    """
    Combine multi-part Altium components into a single component.

    Altium multi-part components can be distinguished by the `_part_count` and
    `_current_part_id` attributes being present, which respectively store the
    total number of parts and the current part number. If that is the case, the
    `_logical_designator` attribute ties together the different parts of the
    component.
    """

    combined_components = []
    multi_part_components_by_designator = {}

    for component in components:
        if "_part_count" in component and "_current_part_id" in component:
            designator = component[LOGICAL_DESIGNATOR]
            multi_part_components_by_designator.setdefault(designator, []).append(component)
        else:
            combined_components.append(component)

    for designator, multi_part_components in multi_part_components_by_designator.items():
        combined_component = multi_part_components[0].copy()
        combined_component[DESIGNATOR_COLUMN_NAME] = designator
        # The combined component shouldn't have the current part id, as it is
        # not any of the parts.
        del combined_component["_current_part_id"]
        combined_components.append(combined_component)

    return combined_components


def _combine_multi_part_components_for_orcad(
    components: list[dict[str, str]],
) -> list[dict[str, str]]:
    """
    Combine multi-part OrCAD components into a single component.

    Multi-part OrCAD components can be distinguished by the "logical_reference"
    attribute, which ties together the different parts of the component.
    """

    combined_components = []
    multi_part_components_by_designator = {}

    for component in components:
        if "_logical_reference" in component:
            designator = component["_logical_reference"]
            multi_part_components_by_designator.setdefault(designator, []).append(component)
        else:
            combined_components.append(component)

    for designator, multi_part_components in multi_part_components_by_designator.items():
        combined_component = multi_part_components[0].copy()
        combined_component[PART_REFERENCE_ATTR_NAME] = designator
        combined_component["_reference"] = designator
        combined_components.append(combined_component)

    return combined_components


def _extract_variations(
    variant: str,
    prjpcb_ini: configparser.ConfigParser,
) -> configparser.SectionProxy:
    """
    Extract the details of a variant from a PrjPcb file.
    """

    available_variants = set()

    for section in prjpcb_ini.sections():
        if section.startswith("ProjectVariant"):
            if prjpcb_ini[section].get("Description") == variant:
                return prjpcb_ini[section]
            else:
                available_variants.add(prjpcb_ini[section].get("Description"))
    raise ValueError(
        f"Variant {variant} not found in PrjPcb file.\n"
        f"Available variants: {', '.join(available_variants)}"
    )


def _apply_variations(
    components: list[dict[str, str]],
    variant_details: configparser.SectionProxy,
    logger: Logger,
) -> list[dict[str, str]]:
    """
    Apply the variations of a specific variant to the components. This should be
    done before the components are mapped to columns or grouped.

    :param components: The components to apply the variations to.
    :param variant_details: The section of the config file dealing with a
        specific variant.

    :returns: The components with the variations applied.
    """

    # List of component UniqueIDs to remove from the BOM.
    components_to_remove: list[str] = []
    # When patching components, the ParamVariation doesn't have the unique ID,
    # only a designator. However, ParamVariations follow the Variation entry, so
    # if we note down the last unique id we saw for a designator when going
    # through the variations, we can use that unique id when handling a param
    # variation. This dict holds that information.
    patch_component_unique_id: dict[str, str] = {}
    # The keys are the unique IDs, and the values are a key-value of the
    # parameter to patch and the value to patch it to.
    components_to_patch: dict[str, list[tuple[str, str]]] = {}

    for key, value in variant_details.items():
        # Note that this is in lowercase, as configparser stores all keys in
        # lowercase.
        if re.match(r"variation[\d+]", key):
            variation_details = dict(details.split("=", 1) for details in value.split("|"))
            try:
                designator = variation_details["Designator"]
                unique_id = variation_details["UniqueId"]
                kind = variation_details["Kind"]
            except KeyError:
                logger.warning(
                    "Designator, UniqueId, or Kind not found in details of variation "
                    f"{variation_details}; skipping this variation."
                )
                continue
            try:
                kind = VariationKind(int(variation_details["Kind"]))
            except ValueError:
                logger.warning(
                    f"Kind {variation_details['Kind']} of variation {variation_details} must be "
                    "either 0, 1 or 2; skipping this variation."
                )
                continue

            if kind == VariationKind.NOT_FITTED:
                components_to_remove.append(unique_id)
            else:
                patch_component_unique_id[designator] = unique_id
        elif re.match(r"paramvariation[\d]+", key):
            variation_id = key.split("paramvariation")[-1]
            designator = variant_details[f"ParamDesignator{variation_id}"]
            variation_details = dict(details.split("=", 1) for details in value.split("|"))
            try:
                unique_id = patch_component_unique_id[designator]
            except KeyError:
                # This can happen sometimes - Altium allows param variations
                # even when the component is not fitted, so we just log and
                # ignore.
                logger.warning(
                    f"ParamVariation{variation_id} found for component {designator} either before "
                    "the corresponding Variation or for a component that is not fitted.\n"
                    "Ignoring this ParamVariation."
                )
                continue

            try:
                parameter_patch = (
                    variation_details["ParameterName"],
                    variation_details["VariantValue"],
                )
            except KeyError:
                logger.warning(
                    f"ParameterName or VariantValue not found in ParamVariation{variation_id} "
                    "details."
                )
                continue

            components_to_patch.setdefault(unique_id, []).append(parameter_patch)

    final_components = []

    for component in components:
        unique_id = component["_unique_id"]
        if unique_id in components_to_remove:
            continue

        if unique_id in components_to_patch:
            new_component = component.copy()
            for parameter, value in components_to_patch[unique_id]:
                new_component[parameter] = value
            final_components.append(new_component)
        else:
            final_components.append(component)

    return final_components


def _filter_blank_components(
    components: list[ComponentAttributes],
    logger: Logger,
) -> list[ComponentAttributes]:
    """
    Remove components that have no attributes, or components for which all
    attributes are empty strings.

    This funtion also debug logs a warning for components that have no attributes.
    """

    final_components = []

    for component in components:
        if not any(component.values()):
            logger.debug(f"Component {component} has no attributes; skipping.")
            continue
        final_components.append(component)

    return final_components


def _find_device_sheet(
    device_sheet: str,
    project_repository: Repository,
    project_file_path: str,
    design_reuse_repos: list[Repository],
    logger: Logger,
) -> tuple[Repository, pathlib.PurePosixPath]:
    """
    Find a device sheet by name in the design reuse repositories.

    This currently does not use the directory structure of the device sheet
    path from the PrjPcb file, and instead only matches on the filename. If
    multiple matches are found, they are logged.

    :param design_reuse_repos: The design reuse repositories to search.
    :param device_sheet: The path of the device sheet as given in the PrjPcb.
    :returns: The first matching repository and the path to the device sheet in
        that repository.
    """

    matches: list[tuple[Repository, pathlib.PurePosixPath]] = []

    # Paths stored in the PrjPcb are Windows paths
    device_sheet_path = pathlib.PureWindowsPath(device_sheet)
    # Note that this will include the .SchDoc extension.
    device_sheet_name = device_sheet_path.name.casefold()

    # First, we'll resolve the actual path of the device sheet in the project
    # repo, and if that's a real file then we use that.
    device_sheet_path_from_repo_root = _resolve_prjpcb_relative_path(
        device_sheet,
        project_file_path,
    ).casefold()
    project_repo_tree = _fetch_all_files_in_repo(project_repository)
    for filepath in project_repo_tree:
        if device_sheet_path_from_repo_root == filepath.casefold():
            logger.info("Found device sheet %s in project repository; using that.", device_sheet)
            return (
                project_repository,
                pathlib.PurePosixPath(filepath),
            )

    for repo in design_reuse_repos:
        files_in_repo = _fetch_all_files_in_repo(repo)
        for file_in_repo in files_in_repo:
            # Paths as reported by ASH are Unix paths
            file_path = pathlib.PurePosixPath(file_in_repo)
            if device_sheet_name == file_path.name.casefold():
                matches.append((repo, file_path))

    if len(matches) == 0:
        raise ValueError(
            f"No matching device sheet found for {device_sheet_path} in the design reuse "
            "repositories.",
        )

    if len(matches) > 1:
        logger.info(
            "Multiple matches found for device sheet %s in design reuse repositories; set log "
            "level to debug to see all matches.",
            device_sheet,
        )
        for match in matches:
            logger.debug("Matching repository: %s, matching file path: %s", match[0].url, match[1])

        first_match = matches[0]
        logger.info(
            "Picking first match, i.e. Repository: %s, File Path: %s",
            first_match[0].url,
            first_match[1],
        )

    return matches[0]


def _fetch_and_parse_annotation_file(
    repository: Repository,
    prjpcb_path: str,
    ref: Ref,
) -> dict:
    """
    Determine the path to the annotation file, fetch it and parse it.

    :param repository: The repository containing the annotation file.
    :param prjpcb_path: The path to the PrjPcb file from the repo root.
    :param ref: The git ref to check.
    :returns: The parsed contents of the annotation file. The returned
        dictionary maps the full component unique id to a list of changes. For
        example, a component with id \\A\\B\\C with channel name CN that should
        change from C1 to C12 will be stored as
        {"\\A\\B\\C": [{"from": "C1", "to": "C12", "channel_name": "CN"}]}.
    """

    # According to the Altium documentation, the annotation file is stored with
    # the PrjPcb file and has the same name as the PrjPcb file, but with a
    # .Annotation extension.

    prjpcb_posix_path = pathlib.PurePosixPath(prjpcb_path)
    prjpcb_directory = prjpcb_posix_path.parent
    annotation_file_path = prjpcb_directory / (prjpcb_posix_path.stem + ".Annotation")
    caseless_annotation_file_path = annotation_file_path.as_posix().casefold()

    # We need to check case-insensitively because the project is made on
    # Windows, but ASH has a Unix-like filesystem.

    files_in_repo = _fetch_all_files_in_repo(repository)
    # Since we're checking case insensitively, we need to make sure there's only
    # one match for the annotation file path.
    matches = []
    for file_path in files_in_repo:
        if file_path.casefold() == caseless_annotation_file_path:
            matches.append(file_path)

    if len(matches) == 0:
        raise Exception(
            f"Could not find annotation file {annotation_file_path} in repository {repository.url}."
        )
    if len(matches) > 1:
        raise Exception(
            f"Multiple matches found for annotation file {annotation_file_path} in repository "
            f"{repository.url} when checking case-insensitively: {matches}."
        )

    annotation_file_path = matches[0]

    annotation_file_contents = repository.get_raw_file(annotation_file_path, ref=ref).decode(
        "utf-8-sig"
    )

    annotation_file_ini = configparser.ConfigParser()
    annotation_file_ini.read_string(annotation_file_contents)

    designator_manager_section = annotation_file_ini["DesignatorManager"]
    changes = {}

    # We'll manually loop through change numbers until we find one which
    # doesn't exist.
    change_number = 0

    while True:
        unique_id_key = f"UniqueID{change_number}"
        if unique_id_key not in designator_manager_section:
            break
        unique_id = designator_manager_section[unique_id_key]
        logical_designator = designator_manager_section[f"LogicalDesignator{change_number}"]
        physical_designator = designator_manager_section[f"PhysicalDesignator{change_number}"]

        if unique_id in changes:
            repository.allspice_client.logger.warning("Multiple changes found for %s", unique_id)
            continue

        changes[unique_id] = {"from": logical_designator, "to": physical_designator}
        change_number += 1

    return changes


def _apply_annotation_file(
    components: list[ComponentAttributes],
    annotations_data: dict,
    unique_ids_mapping: dict[str, str],
    logger: Logger,
) -> list[ComponentAttributes]:
    """
    Apply a Board Level Annotations file to get the compnents with their final
    designators.
    """

    final_components = []

    # Map sheet id to dict of component id to change.
    annotations_by_sheet = {}
    # Map sheet id to dict of designator to list of changes.
    annotations_by_sheet_and_designator = {}

    for unique_id, change in annotations_data.items():
        sheet_id, component_id = unique_id.rsplit("\\", 1)
        annotations_by_sheet.setdefault(sheet_id, {})[component_id] = change
        annotations_by_sheet_and_designator.setdefault(sheet_id, {}).setdefault(
            change["from"], []
        ).append(change)

    for component in components:
        unique_id = component["_unique_id"]
        sheet_id, component_id = unique_id.rsplit("\\", 1)
        ids_to_test = [component_id]
        if component_id in unique_ids_mapping:
            logger.debug("Found component %s in unique ID mapping", component_id)
            other_component_ids = unique_ids_mapping[component_id]
            ids_to_test.extend(other_component_ids)

        annotation_for_component = None
        for id_to_test in ids_to_test:
            annotation_for_component = annotations_by_sheet.get(id_to_test)
            if annotation_for_component:
                break

        if not annotation_for_component:
            logger.debug(
                "No annotation found for component %s by unique ID, trying by designator.",
                component_id,
            )
            component_designator = component[DESIGNATOR_COLUMN_NAME]
            annotations_for_designator = annotations_by_sheet_and_designator.get(sheet_id, {}).get(
                component_designator
            )
            if annotations_for_designator and len(annotations_for_designator) == 1:
                logger.debug(
                    "Found exact match for component %s, using annotation.", component_designator
                )
                annotation_for_component = annotations_for_designator[0]
            else:
                logger.debug(
                    "Found multiple annotations for designator %s; skipping.", component_designator
                )

        if not annotation_for_component:
            final_components.append(component)
            continue

        new_component = component.copy()
        new_component[DESIGNATOR_COLUMN_NAME] = annotation_for_component["to"]
        if LOGICAL_DESIGNATOR in new_component:
            new_component[LOGICAL_DESIGNATOR] = annotation_for_component["to"]
        final_components.append(new_component)
        continue

    return final_components


def _compute_repetitions(components: list[ComponentAttributes]) -> list[ComponentAttributes]:
    """
    In the absence of an Annotation file, manually append suffixes to repeated
    components.
    """

    final_components = []

    # The rough logic here is that for all components with the same ID, i.e.
    # the LAST segment of their unique ID, we sort them by channel name and
    # suffix the index based on its position in the sorted list.

    components_by_id = {}

    for component in components:
        unique_id = component["_unique_id"]
        last_segment = unique_id.split("\\")[-1]
        components_by_id.setdefault(last_segment, []).append(component)

    for components in components_by_id.values():
        if len(components) == 1:
            final_components.append(components[0])
            continue

        # If there are multiple components with the same ID, they *must* be
        # from repeated sheets, so we they *must* have the _channel_name field
        # in the component attributes.
        components.sort(key=lambda component: component["_channel_name"])
        for i, component in enumerate(components):
            new_component = component.copy()
            new_component[DESIGNATOR_COLUMN_NAME] += _letters_for_repetition(i + 1)
            new_component[LOGICAL_DESIGNATOR] += _letters_for_repetition(i + 1)
            final_components.append(new_component)

    return final_components


def _correct_device_sheet_reference_unique_ids(
    hierarchy: SchdocHierarchy,
    prjpcb_ini: configparser.ConfigParser,
    logger: Logger,
) -> SchdocHierarchy:
    """
    When using device sheets, the UniqueIDs of the sheet references in Device
    sheets are updated by Altium to be unique to the project. All UniqueID
    paths to the files are stored in the PrjPcb file, so this function compares
    the hierarchy to the paths in the PrjPcb path to correct the UniqueIDs of
    the SheetRefs to be what they should be in the project.

    This is known to not work in the following cases:

    - There are multiple top sheets in the project. Altium warns in this case,
      so this is not supported.
    - There are multiple sheets with the same name in the project. Altium also
      doesn't support this well and will do unusual things, so we don't
      support this.
    """

    try:
        annotate_section = prjpcb_ini["Annotate"]
    except KeyError:
        logger.warning(
            "Annotate section not found in PrjPcb; Designators of componets within Device Sheets may be incorrect."
        )
        return hierarchy

    paths_to_documents: dict[str, list[str]] = {}
    document_index = 0

    while True:
        try:
            document_name = annotate_section[f"DocumentName{document_index}"]
            unique_id_path = annotate_section[f"UniqueIDPath{document_index}"]
        except KeyError:
            break

        document_index += 1

        if not unique_id_path:
            # The unique ID path can be blank for the top sheet - we'll just
            # skip it, because we already know the top sheet.
            continue

        paths_to_documents.setdefault(document_name, []).append(unique_id_path)

    # How this works:
    #
    # We have a list of paths which are based on the hierarchy of sheet ref
    # unique IDs, and a list of the sheetrefs that are on each sheet. We also
    # know that all instances of a sheet have the same SheetRefs with the same
    # IDs. So if a Sheet1.SchDoc has two children Sheet2.SchDoc and
    # Sheet3.SchDoc, it will have two sheet refs, and across all instances of
    # Sheet1.SchDoc those sheetrefs will have the same unique Id.
    #
    # Therefore, we can take one of the paths to Sheet1, which gives us the
    # path for an instance of Sheet1. Then we take the list of children of
    # Sheet1 and pick a child, say Sheet2, and find all paths to Sheet2 that
    # passes through this instance of Sheet1. We also filter out paths that
    # aren't immediate children of this instance, and so we'll have a list of
    # unique IDs for the sheet refs, which we then reconcile with the sheetrefs
    # we do have.

    changed_hierarchy = {}

    for sheet_name, children in hierarchy.items():
        if not sheet_name.endswith(".SchDoc"):
            sheet_name_with_extension = sheet_name + ".SchDoc"
        else:
            sheet_name_with_extension = sheet_name

        if sheet_name_with_extension not in paths_to_documents:
            changed_hierarchy[sheet_name] = children
            continue

        unique_id_paths = paths_to_documents[sheet_name_with_extension]
        # Since all instances of this sheet should have the same children, we
        # only need one of the unique paths.
        path_to_this_sheet = unique_id_paths[0]
        children_by_name: dict[str, list[AltiumChildSheet]] = {}

        final_children = []
        for child in children:
            if child.kind == AltiumChildSheet.ChildSheetKind.DOCUMENT:
                # We don't need to change the IDs of documents - they're
                # already correct.
                final_children.append(child)
                continue

            children_by_name.setdefault(child.name, []).append(child)

        for child_name, children_of_name in children_by_name.items():
            # The name in the paths_to_documents dict has a SchDoc appended to
            # it
            child_name += ".SchDoc"
            paths_to_children = paths_to_documents.get(child_name, [])

            if not paths_to_children:
                logger.warning(
                    f"Could not find any paths for {child_name} in the PrjPcb file; skipping."
                )
                final_children.extend(children_of_name)
                continue

            ids: set[str] = set()
            for path in paths_to_children:
                if not path.startswith(path_to_this_sheet):
                    continue

                path = path.removeprefix(path_to_this_sheet + "\\")

                if "\\" in path:
                    # This path is deeper in the hierarchy, we can't use it.
                    continue

                # Now the path is the UniqueID of the child sheet
                ids.add(path)

            # Filter out children that already have the correct ids.
            children_to_correct = []
            for child in children_of_name:
                if child.unique_id in ids:
                    ids.remove(child.unique_id)
                    final_children.append(child)
                else:
                    children_to_correct.append(child)

            # Now we have the children who need new ids and the new ids
            for child, unique_id in zip(children_to_correct, ids, strict=True):
                final_children.append(dataclasses.replace(child, unique_id=unique_id))

        changed_hierarchy[sheet_name] = final_children

    return changed_hierarchy


def _create_unique_ids_mapping(prjpcb_ini: configparser.ConfigParser) -> dict:
    """
    If multiple components in the same project have the same UniqueIDs, Altium
    creates a mapping between them. This function reads that mapping from the
    PrjPcb file.
    """

    if "UniqueIdsMappings" not in prjpcb_ini.sections():
        return {}

    unique_ids_mapping = {}
    unique_ids_section = prjpcb_ini["UniqueIdsMappings"]
    # this section starts with one for some reason
    mapping_index = 1

    while True:
        try:
            mapping = unique_ids_section[f"Mapping{mapping_index}"]
        except KeyError:
            break
        (sch_handle, unique_id_mapping) = mapping.split("|")
        _, sch_handle = sch_handle.split("=", 1)
        _, unique_id_mapping = unique_id_mapping.split("=", 1)
        # The sch handle is a path, we only need the last part of it.
        component_id = sch_handle.split("\\")[-1]
        unique_ids_mapping.setdefault(component_id, []).append(unique_id_mapping)
        mapping_index += 1

    return unique_ids_mapping



import base64
import os
from unittest.mock import patch

import pytest
from syrupy.extensions.json import JSONSnapshotExtension

from allspice import AllSpice
from allspice.exceptions import NotYetGeneratedException
from allspice.utils import list_components
from allspice.utils.bom_generation import (
    ColumnConfig,
    generate_bom,
    generate_bom_for_altium,
    generate_bom_for_orcad,
    generate_bom_for_system_capture,
)
from allspice.utils.list_components import list_components_for_altium, list_components_for_orcad
from allspice.utils.netlist_generation import generate_netlist

from .csv_snapshot_extension import CSVSnapshotExtension


@pytest.fixture(scope="session")
def port(pytestconfig):
    """Load --port command-line arg if set"""
    return pytestconfig.getoption("port")


@pytest.fixture(scope="session")
def client_log_level(pytestconfig):
    """Load --client-log-level command-line arg if set"""
    return pytestconfig.getoption("client_log_level")


@pytest.fixture
def instance(port, client_log_level, pytestconfig):
    # The None record mode is the default and is equivalent to "once", but it
    # can also be set to None if recording is disabled entirely.
    if (
        pytestconfig.getoption("record_mode") in ["none", "once", None]
    ) and not pytestconfig.getoption("disable_recording"):
        # If we're using cassettes, we don't want BOM generation to sleep
        # between requests to wait for the generated JSON to be available.
        list_components.SLEEP_FOR_GENERATED_JSON = 0

    if os.environ.get("CI") == "true":
        # The CI runner is anemic and may not be able to generate the outputs
        # in the default number of retries, so we set it to a very high number
        # to make it effectively retry for a fair amount of time if we're not
        # using cassettes. If it cannot generate even in ~100s, that could
        # indicate a real issue.
        list_components.MAX_RETRIES_FOR_GENERATED_JSON = 100

    try:
        g = AllSpice(
            f"http://localhost:{port}",
            open(".token", "r").read().strip(),
            ratelimiting=None,
            log_level=client_log_level or "INFO",
        )
        print("AllSpice Hub Version: " + g.get_version())
        print("API-Token belongs to user: " + g.get_user().username)

        return g
    except Exception:
        assert False, f"AllSpice Hub could not load. Is there: \
                - an Instance running at http://localhost:{port} \
                - a Token at .token \
                    ?"


@pytest.fixture
def setup_for_generation(instance):
    repos = []

    def setup_for_generation_inner(test_name, clone_addr):
        # TODO: we should commit a smaller set of files in this repo so we don't
        #       depend on external data
        nonlocal repos

        instance.requests_post(
            "/repos/migrate",
            data={
                "clone_addr": clone_addr,
                "mirror": False,
                "repo_name": "-".join(["test", test_name]),
                "service": "git",
            },
        )

        repo = instance.get_repository(
            instance.get_user().username,
            "-".join(["test", test_name]),
        )
        repos.append(repo)
        return repo

    yield setup_for_generation_inner

    for repo in repos:
        repo.delete()


@pytest.fixture(scope="module")
def vcr_config():
    return {"filter_headers": ["authorization"]}


@pytest.fixture()
def json_snapshot(snapshot):
    return snapshot.use_extension(JSONSnapshotExtension)


@pytest.fixture()
def csv_snapshot(snapshot):
    return snapshot.use_extension(CSVSnapshotExtension)


@pytest.mark.vcr
def test_bom_generation_flat(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        # We hard-code a ref so that this test is reproducible.
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
    )

    assert len(bom) == 913

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_with_odd_line_endings(
    request, instance, setup_for_generation, csv_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    # We hard-code a ref so that this test is reproducible.
    ref = "95719adde8107958bf40467ee092c45b6ddaba00"
    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }

    new_branch_name = "-".join(["odd-line-endings", request.node.name])
    repo.add_branch(ref, new_branch_name)
    ref = new_branch_name

    files_in_repo = repo.get_git_content(ref=ref)
    prjpcb_file = next((x for x in files_in_repo if x.path == "Archimajor.PrjPcb"), None)
    assert prjpcb_file is not None

    original_prjpcb_sha = prjpcb_file.sha
    prjpcb_content = repo.get_raw_file(prjpcb_file.path, ref=ref).decode("utf-8")
    new_prjpcb_content = prjpcb_content.replace("\r\n", "\n\r")
    new_content_econded = base64.b64encode(new_prjpcb_content.encode("utf-8")).decode("utf-8")
    repo.change_file(
        "Archimajor.PrjPcb",
        original_prjpcb_sha,
        new_content_econded,
        {"branch": ref},
    )

    # Sanity check that the file was changed.
    prjpcb_content_now = repo.get_raw_file("Archimajor.PrjPcb", ref=ref).decode("utf-8")
    assert prjpcb_content_now != prjpcb_content

    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        # Note that ref here is the branch, not a commit sha as in the previous
        # test.
        ref=ref,
    )

    assert len(bom) == 913

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_grouped(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }

    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        group_by=["part_number", "manufacturer", "description"],
        # We hard-code a ref so that this test is reproducible.
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
    )

    assert len(bom) == 108

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_with_folder_hierarchy(
    request,
    instance,
    setup_for_generation,
    csv_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorInFolders.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        group_by=["part_number"],
        # We hard-code a ref so that this test is reproducible.
        ref="e39ecf4de0c191559f5f23478c840ac2b6676d58",
    )

    assert len(bom) == 102
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_with_default_variant(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorVariants.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        ref="916e739f3ad9d956f4e2a293542050e1df9e6f9e",
        # For the variants tests, we don't want to remove non-BOM components
        # because some of them are enabled by the variants, and we want to
        # test that they are included when required.
        remove_non_bom_components=False,
    )

    # Since we haven't specified a variant, this should have the same result
    # as generating a flat BOM. This version of archimajor has a few parts
    # removed even before the variations, so the number of parts is different.
    assert len(bom) == 975

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_with_fitted_variant(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorVariants.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        ref="916e739f3ad9d956f4e2a293542050e1df9e6f9e",
        variant="Fitted",
        remove_non_bom_components=False,
    )

    # Exactly 42 rows should be removed, as that is the number of non-param
    # variations.
    assert len(bom) == 975 - 42

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_with_grouped_variant(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorVariants.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        group_by=["part_number"],
        ref="916e739f3ad9d956f4e2a293542050e1df9e6f9e",
        variant="Fitted",
        remove_non_bom_components=False,
    )

    assert len(bom) == 89

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_altium_with_non_bom_components(
    request, instance, setup_for_generation, csv_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        # We hard-code a ref so that this test is reproducible.
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
        remove_non_bom_components=False,
    )

    assert len(bom) == 1049

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_altium_repeated_multi_part_component(
    request, instance, setup_for_generation, csv_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorRepeated.git",
    )
    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        ref="1bb73a0c862e156557e05876fb268ba086e9d42d",
        remove_non_bom_components=True,
    )

    assert len(bom) == 870

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_altium_with_column_config(
    request, instance, setup_for_generation, csv_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )
    columns = {
        "description": ColumnConfig(
            attributes="PART DESCRIPTION",
            grouped_values_sort=ColumnConfig.SortOrder.DESC,
            grouped_values_allow_duplicates=True,
        ),
        "designator": ColumnConfig(
            attributes="Designator",
            grouped_values_sort=ColumnConfig.SortOrder.DESC,
            grouped_values_separator=";",
        ),
        "manufacturer": ColumnConfig(
            attributes=["Manufacturer", "MANUFACTURER"],
            sort=ColumnConfig.SortOrder.DESC,
        ),
        "tolerance": ColumnConfig(
            attributes=["Tolerance", "TOLERANCE"],
            sort=ColumnConfig.SortOrder.ASC,
            remove_rows_matching="-NA-",
            skip_in_output=True,
        ),
        "part_number": ColumnConfig(
            attributes=["PART", "MANUFACTURER #"],
            sort=ColumnConfig.SortOrder.ASC,
        ),
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        columns,
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
        group_by=["part_number"],
    )

    assert len(bom) == 64
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_altium_repeated_multi_part_component_variant(
    request, instance, setup_for_generation, csv_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorRepeatedVariant.git",
    )
    attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        attributes_mapping,
        ref="3f8ddd6b5161aebc61a3ed87b665ba0a64cc6e89",
        variant="Fitted",
        remove_non_bom_components=True,
    )

    assert len(bom) == 869

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_altium_with_device_sheets(
    request,
    instance,
    setup_for_generation,
    csv_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/Altium-Device-Sheet-Usage-Demo",
    )
    design_reuse_repo = setup_for_generation(
        request.node.name + "_reuse",
        "https://hub.allspice.io/NoIndexTests/Altium-Device-Sheets",
    )
    attributes_mapping = {
        "Name": ["_name"],
        "Designator": ["Designator"],
        "Comment": ["Comment"],
    }
    bom = generate_bom_for_altium(
        instance,
        repo,
        "DCDC Regulators Breakout/DCDC Regulators Breakout.PrjPcb",
        attributes_mapping,
        group_by=["Comment"],
        design_reuse_repos=[design_reuse_repo],
        ref="5f2bdd30f57eb8ea6699dc9dcb098bc34d60f7a3",
    )

    assert len(bom) == 13
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_altium_with_hierarchical_device_sheets(
    request,
    instance,
    setup_for_generation,
    csv_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/Altium-Hierarchical-Device-Sheet-Repetitions-Demo",
    )
    design_reuse_repo = setup_for_generation(
        request.node.name + "_reuse",
        "https://hub.allspice.io/NoIndexTests/Altium-Device-Sheets-Hierarchical-Repetitions",
    )

    attributes_mapping = {
        "Name": ["_name"],
        "Designator": ColumnConfig(
            attributes=["Designator"],
            grouped_values_sort=ColumnConfig.SortOrder.ASC,
        ),
        "Comment": ColumnConfig(attributes=["Comment"], sort=ColumnConfig.SortOrder.ASC),
    }

    bom = generate_bom_for_altium(
        instance,
        repo,
        "NestedDeviceSheets.PrjPcb",
        attributes_mapping,
        group_by=["Comment"],
        design_reuse_repos=[design_reuse_repo],
    )

    assert len(bom) == 14
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_orcad(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/beagleplay.git",
    )

    attributes_mapping = {
        "Name": ["_name"],
        "Description": "Description",
        "Reference designator": ["Part Reference"],
        "Manufacturer": ["Manufacturer", "MANUFACTURER"],
        "Part Number": ["Manufacturer PN", "PN"],
    }

    bom = generate_bom_for_orcad(
        instance,
        repo,
        "Design/BEAGLEPLAYV10_221227.DSN",
        attributes_mapping,
        # We hard-code a ref so that this test is reproducible.
        ref="7a59a98ae27dc4fd9e2bd8975ff90cdb44a366ea",
    )

    assert len(bom) == 846

    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_orcad_with_column_config(
    request, instance, setup_for_generation, csv_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/AllSpiceMirrors/beagleplay.git",
    )
    columns = {
        "Manufacturer": ColumnConfig(
            attributes=["Manufacturer", "MANUFACTURER"],
            sort=ColumnConfig.SortOrder.ASC,
        ),
        "Reference designator": "Part Reference",
        "Name": ColumnConfig(
            attributes="_name",
            grouped_values_allow_duplicates=False,
            grouped_values_sort=ColumnConfig.SortOrder.ASC,
        ),
        "Description": ColumnConfig(
            attributes="Description",
            grouped_values_sort=ColumnConfig.SortOrder.DESC,
            grouped_values_separator=";",
        ),
        "Part Number": ColumnConfig(
            attributes=["Manufacturer PN", "PN"],
            remove_rows_matching="^TP",
        ),
    }
    bom = generate_bom_for_orcad(
        instance,
        repo,
        "Design/BEAGLEPLAYV10_221227.DSN",
        columns,
        ref="7a59a98ae27dc4fd9e2bd8975ff90cdb44a366ea",
    )

    assert len(bom) == 777
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_system_capture(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )

    attributes_mapping = {
        "Description": "VALUE",
        "Designator": ["LOCATION"],
        "Part Number": ["VENDOR_PN", "PN"],
    }
    bom = generate_bom_for_system_capture(
        instance,
        repo,
        "parallella_schematic.sdax",
        attributes_mapping,
        # We hard-code a ref so that this test is reproducible.
        ref="e03461e6bbe72f10b163462cf9325b0309e87201",
    )

    assert len(bom) == 564
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_bom_generation_system_capture_grouped_failure(request, instance, setup_for_generation):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )
    with pytest.raises(ValueError, match="Group by column Name not found in selected columns"):
        generate_bom_for_system_capture(
            instance,
            repo,
            "parallella_schematic.sdax",
            {},
            group_by=["Name"],
        )


# The following tests are for the generate_bom function, which is a wrapper
# around the more specific functions for each EDA tool. We test the specific
# functions above, so these tests are just to make sure the wrapper works as
# expected.
@pytest.mark.vcr
def test_generate_bom_altium(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    altium_attributes_mapping = {
        "description": ["PART DESCRIPTION"],
        "designator": ["Designator"],
        "manufacturer": ["Manufacturer", "MANUFACTURER"],
        "part_number": ["PART", "MANUFACTURER #"],
    }
    bom = generate_bom(
        instance,
        repo,
        "Archimajor.PrjPcb",
        altium_attributes_mapping,
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
    )
    assert len(bom) == 913
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_generate_bom_orcad(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/beagleplay.git",
    )
    orcad_attributes_mapping = {
        "Name": ["_name"],
        "Description": "Description",
        "Reference designator": ["Part Reference"],
        "Manufacturer": ["Manufacturer", "MANUFACTURER"],
        "Part Number": ["Manufacturer PN", "PN"],
    }
    bom = generate_bom(
        instance,
        repo,
        "Design/BEAGLEPLAYV10_221227.DSN",
        orcad_attributes_mapping,
        ref="7a59a98ae27dc4fd9e2bd8975ff90cdb44a366ea",
    )
    assert len(bom) == 846
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_generate_bom_system_capture(request, instance, setup_for_generation, csv_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )
    system_capture_attributes_mapping = {
        "Description": "VALUE",
        "Designator": ["LOCATION"],
        "Part Number": ["VENDOR_PN", "PN"],
    }
    bom = generate_bom(
        instance,
        repo,
        "parallella_schematic.sdax",
        system_capture_attributes_mapping,
        ref="e03461e6bbe72f10b163462cf9325b0309e87201",
    )
    assert len(bom) == 564
    assert bom == csv_snapshot


@pytest.mark.vcr
def test_generate_bom_system_capture_fails_with_variant(request, instance, setup_for_generation):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )
    with pytest.raises(ValueError, match="Variant is not supported for System"):
        generate_bom(
            instance,
            repo,
            "parallella_schematic.sdax",
            {},
            variant="Fitted",
        )


@pytest.mark.vcr
def test_orcad_components_list(request, instance, setup_for_generation, json_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/beagleplay.git",
    )

    components = list_components_for_orcad(
        instance,
        repo,
        "Design/BEAGLEPLAYV10_221227.DSN",
        # We hard-code a ref so that this test is reproducible.
        ref="7a59a98ae27dc4fd9e2bd8975ff90cdb44a366ea",
    )

    assert len(components) == 870
    assert components == json_snapshot


@pytest.mark.vcr
def test_altium_components_list(request, instance, setup_for_generation, json_snapshot):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    components = list_components_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        # We hard-code a ref so that this test is reproducible.
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
    )

    assert len(components) == 1061
    assert components == json_snapshot


@pytest.mark.vcr
def test_altium_components_list_with_folder_hierarchy(
    request,
    instance,
    setup_for_generation,
    json_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorInFolders.git",
    )

    components = list_components_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        # We hard-code a ref so that this test is reproducible.
        ref="e39ecf4de0c191559f5f23478c840ac2b6676d58",
    )

    assert len(components) == 1049
    assert components == json_snapshot


@pytest.mark.vcr
def test_altium_components_list_with_fitted_variant(
    request,
    instance,
    setup_for_generation,
    json_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorVariants.git",
    )

    components = list_components_for_altium(
        instance,
        repo,
        "Archimajor.PrjPcb",
        # We hard-code a ref so that this test is reproducible.
        ref="916e739f3ad9d956f4e2a293542050e1df9e6f9e",
        variant="Fitted",
    )

    assert len(components) == 945
    assert components == json_snapshot


@pytest.mark.vcr
def test_altium_components_list_with_device_sheets(
    request,
    instance,
    setup_for_generation,
    json_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/Altium-Device-Sheet-Usage-Demo",
    )
    design_reuse_repo = setup_for_generation(
        request.node.name + "_reuse",
        "https://hub.allspice.io/NoIndexTests/Altium-Device-Sheets",
    )
    components = list_components_for_altium(
        instance,
        repo,
        "DCDC Regulators Breakout/DCDC Regulators Breakout.PrjPcb",
        design_reuse_repos=[design_reuse_repo],
        ref="5f2bdd30f57eb8ea6699dc9dcb098bc34d60f7a3",
    )

    assert len(components) == 38
    assert components == json_snapshot


@pytest.mark.vcr
def test_altium_components_list_with_annotations(
    request, instance, setup_for_generation, json_snapshot
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/FlatSat",
    )
    components = list_components_for_altium(
        instance,
        repo,
        "FlatSat/FlatSat.PrjPCB",
        ref="471d42ba87032682c7dc7a0235ffcc02808a3e37",
    )

    assert len(components) == 408
    assert components == json_snapshot


@pytest.mark.vcr
def test_altium_components_list_with_hierarchical_device_sheets_and_annotations(
    request,
    instance,
    setup_for_generation,
    json_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/Altium-Hierarchical-Device-Sheet-Repetitions-Demo",
    )
    design_reuse_repo = setup_for_generation(
        request.node.name + "_reuse",
        "https://hub.allspice.io/NoIndexTests/Altium-Device-Sheets-Hierarchical-Repetitions",
    )

    components = list_components_for_altium(
        instance,
        repo,
        "NestedDeviceSheets.PrjPcb",
        design_reuse_repos=[design_reuse_repo],
    )

    components.sort(key=lambda x: x["Designator"])
    assert len(components) == 980
    assert components == json_snapshot


@pytest.mark.vcr
def test_system_capture_components_list(
    request,
    instance,
    setup_for_generation,
    json_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )
    components = list_components.list_components_for_system_capture(
        instance,
        repo,
        "parallella_schematic.sdax",
        # We hard-code a ref so that this test is reproducible.
        ref="e03461e6bbe72f10b163462cf9325b0309e87201",
    )
    assert len(components) == 564
    assert components == json_snapshot


# The following tests are for the list_components function, which is a wrapper
# around the more specific functions for each EDA tool. We test the specific
# functions above, so these tests are just to make sure the wrapper works as
# expected.


@pytest.mark.vcr
def test_list_components_system_capture(
    request,
    instance,
    setup_for_generation,
    json_snapshot,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )
    components = list_components.list_components(
        instance,
        repo,
        "parallella_schematic.sdax",
        # We hard-code a ref so that this test is reproducible.
        ref="e03461e6bbe72f10b163462cf9325b0309e87201",
    )
    assert len(components) == 564
    assert components == json_snapshot


@pytest.mark.vcr
def test_list_components_system_capture_fails_with_variant(
    request,
    instance,
    setup_for_generation,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )
    with pytest.raises(ValueError, match="Variant is not supported for System"):
        list_components.list_components(
            instance,
            repo,
            "parallella_schematic.sdax",
            variant="Fitted",
        )


def test_list_components_retries_time_out(
    request,
    instance,
    setup_for_generation,
):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/parallela-sdax.git",
    )

    with patch.object(repo, "get_generated_json", side_effect=NotYetGeneratedException):
        with pytest.raises(TimeoutError):
            list_components.list_components(
                instance,
                repo,
                "parallella_schematic.sdax",
            )


@pytest.mark.vcr
def test_netlist_generation(request, instance, setup_for_generation):
    repo = setup_for_generation(
        request.node.name,
        "https://hub.allspice.io/NoIndexTests/ArchimajorDemo.git",
    )

    netlist = generate_netlist(
        instance,
        repo,
        "Archimajor.PcbDoc",
        # We hard-code a ref so that this test is reproducible.
        ref="95719adde8107958bf40467ee092c45b6ddaba00",
    )
    assert len(netlist) == 682

    nets = list(netlist.keys())

    nets.sort()

    with open("tests/data/archimajor_netlist_expected.net", "r") as f:
        for net in nets:
            assert (net + "\n") == f.readline()
            pins_on_net = sorted(netlist[net])
            assert (" " + " ".join(pins_on_net) + "\n") == f.readline()



import base64
import datetime
import time
import uuid

import pytest

from allspice import (
    AllSpice,
    Branch,
    Comment,
    DesignReview,
    Issue,
    Milestone,
    NotFoundException,
    Organization,
    Repository,
    Team,
    User,
)
from allspice.apiobject import CommitStatusState, Util
from allspice.exceptions import NotYetGeneratedException

# put a ".token" file into your directory containg only the token for AllSpice Hub


@pytest.fixture(scope="session")
def port(pytestconfig):
    """Load --port command-line arg if set"""
    return pytestconfig.getoption("port")


@pytest.fixture
def instance(port, scope="module"):
    try:
        g = AllSpice(
            f"http://localhost:{port}",
            open(".token", "r").read().strip(),
            ratelimiting=None,
        )
        print("AllSpice Hub Version: " + g.get_version())
        print("API-Token belongs to user: " + g.get_user().username)
        return g
    except Exception:
        assert False, f"AllSpice Hub could not load. \
                - Instance running at http://localhost:{port} \
                - Token at .token   \
                    ?"


# make up some fresh names for the tests run
test_org = "org_" + uuid.uuid4().hex[:8]
test_user = "user_" + uuid.uuid4().hex[:8]
test_team = "team_" + uuid.uuid4().hex[:8]  # team names seem to have a rather low max lenght
test_repo = "repo_" + uuid.uuid4().hex[:8]
# Topic names can't have underscores.
test_topic = "topic-" + uuid.uuid4().hex[:8]


def test_token_owner(instance):
    user = instance.get_user()
    assert user.username == "test", "Token user not 'tests'."
    assert user.is_admin, "Testuser is not Admin - Tests may fail"


def test_allspice_hub_version(instance):
    assert instance.get_version().startswith("1."), "No Version String returned"


def test_fail_get_non_existent_user(instance):
    with pytest.raises(NotFoundException):
        User.request(instance, test_user)


def test_fail_get_non_existent_org(instance):
    with pytest.raises(NotFoundException):
        Organization.request(instance, test_org)


def test_fail_get_non_existent_repo(instance):
    with pytest.raises(NotFoundException):
        Repository.request(instance, test_user, test_repo)


def test_create_user(instance):
    email = test_user + "@example.org"
    user = instance.create_user(test_user, email, "abcdefg1.23AB", send_notify=False)
    assert user.username == test_user
    assert user.login == test_user
    assert email in user.emails
    assert user.email == email
    assert not user.is_admin
    assert isinstance(user.id, int)
    assert user.is_admin is False


def test_change_user(instance):
    user = instance.get_user_by_name(test_user)
    location = "a house"
    user.location = location
    new_fullname = "Other Test Full Name"
    user.full_name = new_fullname
    user.commit(user.username, 0)
    del user
    user = instance.get_user_by_name(test_user)
    assert user.full_name == new_fullname
    assert user.location == location


def test_create_org(instance):
    user = instance.get_user()
    org = instance.create_org(user, test_org, "some-desc", "loc")
    assert org.get_members()[0] == user
    assert org.description == "some-desc"
    assert org.username == test_org
    assert org.location == "loc"
    assert not org.website
    assert not org.full_name


def test_non_changable_field(instance):
    org = Organization.request(instance, test_org)
    with pytest.raises(AttributeError):
        org.id = 55


def test_create_repo_userowned(instance):
    org = User.request(instance, test_user)
    repo = instance.create_repo(org, test_repo, "user owned repo")
    assert repo.description == "user owned repo"
    assert repo.owner == org
    assert repo.name == test_repo
    assert not repo.private


def test_edit_org_fields_and_commit(instance):
    org = Organization.request(instance, test_org)
    org.description = "some thing other man"
    org.location = "somewehre new"
    org.visibility = "public"
    org.website = "http:\\\\testurl.com"
    org.commit()
    org2 = Organization.request(instance, test_org)
    assert org2.name == test_org
    assert org2.description == "some thing other man"
    assert org2.location == "somewehre new"
    # assert org2.visibility == "private" # after commiting, this field just vanishes (Bug?)
    assert org2.website == "http:\\\\testurl.com"


def test_create_repo_orgowned(instance):
    org = Organization.request(instance, test_org)
    repo = instance.create_repo(org, test_repo, "descr")
    assert repo.description == "descr"
    assert repo.owner == org
    assert repo.name == test_repo
    assert not repo.private


def test_get_repository(instance):
    repo = instance.get_repository(test_org, test_repo)
    assert repo is not None
    assert repo.name == test_repo
    assert repo.owner.username == test_org


def test_add_content_to_repo(instance):
    repo = Repository.request(instance, test_org, test_repo)
    file_content = open("tests/data/test.pcbdoc", "rb").read()
    file_content = base64.b64encode(file_content).decode("utf-8")
    repo.create_file("test.pcbdoc", file_content)
    assert len(repo.get_commits()) == 2
    assert [content.name for content in repo.get_git_content()] == [
        "README.md",
        "test.pcbdoc",
    ]


def test_get_repo_tree(instance):
    repo = Repository.request(instance, test_org, test_repo)
    tree = repo.get_tree()
    assert len(tree) == 2
    assert tree[0].path == "README.md"


def test_get_json_before_generated(instance):
    repo = Repository.request(instance, test_org, test_repo)
    with pytest.raises(NotYetGeneratedException):
        repo.get_generated_json("test.pcbdoc")


def test_get_svg_before_generated(instance):
    repo = Repository.request(instance, test_org, test_repo)
    with pytest.raises(NotYetGeneratedException):
        repo.get_generated_svg("test.pcbdoc")


def test_get_generated_json(instance):
    repo = Repository.request(instance, test_org, test_repo)
    branch = repo.get_branches()[0]
    # This is scuffed but not much better we can do :(
    while True:
        try:
            json = repo.get_generated_json("test.pcbdoc", branch)
            break
        except NotYetGeneratedException:
            time.sleep(1)
            pass
    assert json is not None
    assert json["type"] == "Pcb"


def test_get_generated_svg(instance):
    repo = Repository.request(instance, test_org, test_repo)
    branch = repo.get_branches()[0]
    while True:
        try:
            svg = repo.get_generated_svg("test.pcbdoc", branch)
            break
        except NotYetGeneratedException:
            time.sleep(1)
            pass
    assert svg is not None
    assert svg.startswith(b"<svg")


def test_get_repository_non_existent(instance):
    with pytest.raises(NotFoundException):
        instance.get_repository("doesnotexist", "doesnotexist")


def test_repo_topics(instance):
    # Since topics aren't part of the Repository object directly, we
    # have to test both get and add at the same time.
    repo = Repository.request(instance, test_org, test_repo)
    topics = repo.get_topics()
    assert len(topics) == 0

    repo.add_topic(test_topic)
    topics = repo.get_topics()
    assert len(topics) == 1
    assert topics[0] == test_topic


def test_search_repos(instance):
    repos = Repository.search(instance, test_repo)
    # Two repos have been made with this name so far!
    assert len(repos) == 2
    assert repos[0].name == test_repo

    repos = Repository.search(instance, test_topic, topic=True)
    assert len(repos) == 1
    assert repos[0].name == test_repo


def test_patch_repo(instance):
    fields = {
        "allow_rebase": False,
        "description": "new description",
        "has_projects": True,
        "private": True,
    }
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    for field, value in fields.items():
        setattr(repo, field, value)
    repo.commit()
    repo = org.get_repository(test_repo)
    for field, value in fields.items():
        assert getattr(repo, field) == value


def test_list_branches(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    branches = repo.get_branches()
    assert len(branches) > 0
    master = [b for b in branches if b.name == "master"]
    assert len(master) > 0


def test_get_branch(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    branch = repo.get_branch("master")
    assert branch is not None
    assert branch.name == "master"


def test_get_branch_non_existent(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    with pytest.raises(NotFoundException):
        repo.get_branch("doesnotexist")


def test_list_files_and_content(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    content = repo.get_git_content()
    # Readme file should exist in any new repo
    #  content is the description given during creation
    readmes = [c for c in content if c.name == "README.md"]
    assert len(readmes) > 0
    readme_content = repo.get_file_content(readmes[0])
    assert len(readme_content) > 0
    assert "descr" in str(base64.b64decode(readme_content))


def test_get_raw_file(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    readme_content = repo.get_raw_file("README.md")
    assert len(readme_content) > 0
    assert "descr" in str(readme_content)


def test_repo_download_file(instance, tmp_path):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    filename = uuid.uuid4().hex[:8] + ".md"
    filepath = tmp_path / filename
    with open(filepath, "wb") as f:
        repo.download_to_file("README.md", f)
    with open(filepath, "rb") as f:
        readme_content = f.read().decode("utf-8")
        assert len(readme_content) > 0
        assert "descr" in readme_content


def test_create_file(instance):
    TESTFILE_CONENTE = "TestStringFileContent"
    TESTFILE_CONENTE_B64 = base64.b64encode(bytes(TESTFILE_CONENTE, "utf-8"))
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    repo.create_file("testfile.md", content=TESTFILE_CONENTE_B64.decode("ascii"))
    # test if putting was successful
    content = repo.get_git_content()
    readmes = [c for c in content if c.name == "testfile.md"]
    assert len(readmes) > 0
    readme_content = repo.get_file_content(readmes[0])
    assert len(readme_content) > 0
    assert TESTFILE_CONENTE in str(base64.b64decode(readme_content))


def test_change_file(instance):
    TESTFILE_CONENTE = "TestStringFileContent with changed content now"
    TESTFILE_CONENTE_B64 = base64.b64encode(bytes(TESTFILE_CONENTE, "utf-8"))
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    # figure out the sha of the file to change
    content = repo.get_git_content()
    readmes = [c for c in content if c.name == "testfile.md"]
    # change
    repo.change_file("testfile.md", readmes[0].sha, content=TESTFILE_CONENTE_B64.decode("ascii"))
    # test if putting was successful
    content = repo.get_git_content()
    readmes = [c for c in content if c.name == "testfile.md"]
    assert len(readmes) > 0
    readme_content = repo.get_file_content(readmes[0])
    assert len(readme_content) > 0
    assert TESTFILE_CONENTE in str(base64.b64decode(readme_content))


def test_delete_file(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    # figure out the sha of the file to change
    content = repo.get_git_content()
    readmes = [c for c in content if c.name == "testfile.md"]
    # delete
    repo.delete_file("testfile.md", readmes[0].sha)
    # test if deletion was successful
    content = repo.get_git_content()
    readmes = [c for c in content if c.name == "testfile.md"]
    assert len(readmes) == 0


def test_create_branch(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    branches = repo.get_branches()
    master = [b for b in branches if b.name == "master"]
    assert len(master) > 0
    branch = repo.add_branch(master[0], "test_branch")
    assert branch.name == "test_branch"
    branch = repo.get_branch("test_branch")
    assert branch.name == "test_branch"


def test_create_branch_from_str_ref(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    new_branch_name = "branch-" + uuid.uuid4().hex[:8]
    branch = repo.add_branch("master", new_branch_name)
    assert branch.name == new_branch_name
    branch = repo.get_branch(new_branch_name)
    assert branch.name == new_branch_name


def test_create_team(instance):
    org = Organization.request(instance, test_org)
    team = instance.create_team(org, test_team, "descr")
    assert team.name == test_team
    assert team.description == "descr"
    assert team.organization == org


def test_add_repo_to_team(instance):
    org = Organization.request(instance, test_org)
    team = org.get_team(test_team)
    repository = Repository.request(instance, test_org, test_repo)

    # First with name
    team.add_repo(org, test_repo)
    assert test_repo in [repo.name for repo in team.get_repos()]

    team.delete()

    team = instance.create_team(org, test_team, "descr")
    # Then with object
    team.add_repo(org, repository)
    assert test_repo in [repo.name for repo in team.get_repos()]


def test_create_team_without_units_map(instance):
    org = Organization.request(instance, test_org)
    team = instance.create_team(org, test_team + "1", "descr")
    permission = team.permission
    assert set(team.units_map.keys()) == set(team.units)
    assert list(team.units_map.values()) == [permission] * len(team.units)


def test_create_team_with_units_map(instance):
    org = Organization.request(instance, test_org)
    team = instance.create_team(
        org,
        test_team + "2",
        "descr",
        units_map={"repo.code": "write", "repo.wiki": "admin"},
    )
    assert set(team.units) == set(["repo.code", "repo.wiki"])
    assert team.units_map == {"repo.code": "write", "repo.wiki": "admin"}


def test_patch_team(instance):
    fields = {
        "can_create_org_repo": True,
        "description": "patched description",
        "includes_all_repositories": True,
        "name": "newname",
        "permission": "write",
    }
    org = Organization.request(instance, test_org)
    team = instance.create_team(org, test_team[:1], "descr")
    for field, value in fields.items():
        setattr(team, field, value)
    team.commit()
    team = Team.request(instance, team.id)
    for field, value in fields.items():
        assert getattr(team, field) == value


def test_request_team(instance):
    org = Organization.request(instance, test_org)
    team = org.get_team(test_team)
    team2 = Team.request(instance, team.id)
    assert team.name == team2.name


def test_create_milestone(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repository(test_repo)
    ms = repo.create_milestone("I love this Milestone", "Find an otter to adopt this milestone")
    assert isinstance(ms, Milestone)
    assert ms.title == "I love this Milestone"


def test_user_teams(instance):
    org = Organization.request(instance, test_org)
    team = org.get_team(test_team)
    user = instance.get_user_by_name(test_user)

    team.add_user(user)
    teams = user.get_teams()
    assert team in teams
    team_members = team.get_members()
    assert user in team_members

    team.remove_team_member(user.login)
    teams = user.get_teams()
    assert team not in teams
    team_members = team.get_members()
    assert user not in team_members


def test_get_accessible_repositories(instance):
    user = instance.get_user_by_name(test_user)
    repos = user.get_accessible_repos()
    assert len(repos) > 0


def test_create_issue(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = Issue.create_issue(instance, repo, "TestIssue", "Body text with this issue")
    assert issue.state == Issue.OPENED
    assert issue.title == "TestIssue"
    assert issue.body == "Body text with this issue"


def test_hashing(instance):
    # just call the hash function of each object to see if something bad happens
    org = Organization.request(instance, test_org)
    team = org.get_team(test_team)
    user = instance.get_user_by_name(test_user)
    # TODO test for milestones (Todo: add milestone adding)
    repo = org.get_repositories()[0]
    milestone = repo.create_milestone("mystone", "this is only a teststone")
    issue = repo.get_issues()[0]
    branch = repo.get_branches()[0]
    commit = repo.get_commits()[0]
    assert len(set([org, team, user, repo, issue, branch, commit, milestone]))


def test_change_issue(instance):
    org = Organization.request(instance, test_org)
    repo = org.get_repositories()[0]
    ms_title = "othermilestone"
    issue = Issue.create_issue(instance, repo, "IssueTestissue with Testinput", "asdf2332")
    new_body = "some new description with some more of that char stuff :)"
    issue.body = new_body
    issue.commit()
    number = issue.number
    del issue
    issue2 = Issue.request(instance, org.username, repo.name, number)
    assert issue2.body == new_body
    milestone = repo.create_milestone(ms_title, "this is only a teststone2")
    issue2.milestone = milestone
    issue2.commit()
    del issue2
    issue3 = Issue.request(instance, org.username, repo.name, number)
    assert issue3.milestone is not None
    assert issue3.milestone.description == "this is only a teststone2"
    issues = repo.get_issues()
    assert (
        len(
            [
                issue
                for issue in issues
                if issue.milestone is not None and issue.milestone.title == ms_title
            ]
        )
        > 0
    )


def test_create_issue_attachment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    attachment = issue.create_attachment(open("requirements.txt", "rb"))
    assert attachment.name == "requirements.txt"
    assert attachment.download_count == 0


def test_get_issue_attachments(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    attachments = issue.get_attachments()
    assert len(attachments) > 0
    assert attachments[0].name == "requirements.txt"

    issue = Issue.request(instance, org.username, repo.name, issue.number)
    assert len(issue.assets) == 1
    assert issue.assets[0].name == "requirements.txt"


def test_create_issue_comment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.create_comment("this is a comment")
    assert comment.body == "this is a comment"
    assert comment.user.username == "test"


def test_get_issue_comments(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comments = issue.get_comments()
    assert len(comments) > 0
    assert comments[0].body == "this is a comment"


def test_edit_issue_comment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.get_comments()[0]
    comment.body = "this is a new comment"
    comment.commit()
    assert comment.body == "this is a new comment"
    comment_id = comment.id
    comment2 = Comment.request(instance, org.username, repo.name, comment_id)
    assert comment2.body == "this is a new comment"


def test_delete_issue_comment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.get_comments()[0]
    comment_id = comment.id
    comment.delete()
    with pytest.raises(NotFoundException) as _:
        Comment.request(instance, org.username, repo.name, comment_id)


def test_create_issue_comment_attachment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.create_comment("this is a comment that will have an attachment")
    attachment = comment.create_attachment(open("requirements.txt", "rb"))
    assert attachment.name == "requirements.txt"
    assert attachment.download_count == 0


def test_create_issue_comment_attachment_with_name(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.create_comment("this is a comment that will have an attachment")
    attachment = comment.create_attachment(open("requirements.txt", "rb"), "something else.txt")
    assert attachment.name == "something else.txt"
    assert attachment.download_count == 0


def test_get_issue_comment_attachments(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.get_comments()[0]
    attachments = comment.get_attachments()
    assert len(attachments) > 0
    assert attachments[0].name == "requirements.txt"


def test_download_issue_comment_attachment(instance, tmp_path):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.get_comments()[0]
    attachment = comment.get_attachments()[0]

    filename = uuid.uuid4().hex[:8] + ".txt"
    filepath = tmp_path / filename
    with open(filepath, "wb") as f:
        attachment.download_to_file(f)

    with open(filepath, "r") as actual_f:
        with open("requirements.txt", "r") as expected_f:
            attachment_content = actual_f.read()
            assert len(attachment_content) > 0

            expected_content = expected_f.read()
            assert expected_content == attachment_content


def test_edit_issue_comment_attachment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.get_comments()[0]
    attachment = comment.get_attachments()[0]
    comment.edit_attachment(attachment, {"name": "this is a new attachment"})
    del attachment
    attachment2 = comment.get_attachments()[0]
    assert attachment2.name == "this is a new attachment"


def test_delete_issue_comment_attachment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    issue = repo.get_issues()[0]
    comment = issue.get_comments()[0]
    attachment = comment.get_attachments()[0]
    comment.delete_attachment(attachment)
    assert len(comment.get_attachments()) == 0


def test_create_design_review(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    branch = Branch.request(instance, org.username, test_repo, "master")
    repo.create_file(
        "new_file.txt",
        base64.b64encode(b"new file contents").decode("utf-8"),
        {"branch": "test_branch"},
    )
    due_date = datetime.datetime.now() + datetime.timedelta(days=7)
    review = repo.create_design_review(
        "TestDesignReview",
        "test_branch",
        branch,
        body="This is a test review",
        assignees=["test"],
        due_date=due_date,
    )

    assert review.state == DesignReview.OPEN
    assert review.title == "TestDesignReview"
    assert review.base == "master"
    assert review.head == "test_branch"
    assert review.body == "This is a test review"
    assert review.assignees[0].username == "test"

    review_due_date = Util.convert_time(review.due_date)
    assert review_due_date.date() == due_date.date()


def test_get_design_reviews(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]

    assert dr.title == "TestDesignReview"
    assert dr.base == "master"
    assert dr.head == "test_branch"
    assert dr.body == "This is a test review"


def test_edit_design_review(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]
    new_branch = repo.add_branch(repo.get_branch("master"), "test_branch2")
    dr.title = "TestDesignReview2"
    dr.body = "This is a test review2"
    dr.due_date = None
    dr.base = new_branch
    dr.commit()
    del dr
    dr = repo.get_design_reviews()[0]
    assert dr.title == "TestDesignReview2"
    assert dr.base == "test_branch2"
    assert dr.head == "test_branch"
    assert dr.body == "This is a test review2"
    assert dr.due_date is None


def test_create_design_review_comment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]
    comment = dr.create_comment("This is a test comment")
    assert comment.body == "This is a test comment"


def test_get_design_review_comments(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]
    comments = dr.get_comments()
    assert len(comments) > 0
    assert comments[0].body == "This is a test comment"


def test_add_design_review_comment_attachment(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]
    comment = dr.get_comments()[0]
    attachment = comment.create_attachment(open("requirements.txt", "rb"))
    assert attachment.name == "requirements.txt"
    assert attachment.download_count == 0


def test_get_design_review_attachments(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]
    comment = dr.get_comments()[0]
    attachments = comment.get_attachments()
    assert len(attachments) > 0
    assert attachments[0].name == "requirements.txt"


def test_merge_design_review(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    dr = repo.get_design_reviews()[0]

    assert dr.state == "open"

    # The API may fail with this message. It isn't in the scope of this test to
    # make sure the API can merge, so we'll keep retrying until it works or we
    # retry too many times.
    attempts = 0
    while attempts < 5:
        try:
            dr.merge(DesignReview.MergeType.MERGE)
            break
        except Exception as e:
            if "Please try again later" in str(e):
                attempts += 1
                time.sleep(1)
                continue
            else:
                raise e

    dr = repo.get_design_reviews()[0]
    assert dr.state == "closed"


def test_repo_create_release(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    # Just a tag should be enough.
    release = repo.create_release("v0.0.1")
    assert release.tag_name == "v0.0.1"
    release = repo.create_release("v0.1.0", "v0.1.0 release", "release with new tag")
    assert release.tag_name == "v0.1.0"
    assert release.name == "v0.1.0 release"
    assert release.body == "release with new tag"


def test_get_repo_releases(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    releases = repo.get_releases()
    assert len(releases) > 0
    assert releases[0].name == "v0.1.0 release"


def test_get_repo_latest_release(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    assert release is not None
    assert release.name == "v0.1.0 release"


def test_get_release_by_tag(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_release_by_tag("v0.1.0")
    assert release is not None
    assert release.name == "v0.1.0 release"


def test_edit_release(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    # Note that the tag hasn't changed
    release.name = "v0.1.1 release"
    release.body = "release with changed name"
    release.commit()
    del release
    release = repo.get_latest_release()
    assert release.name == "v0.1.1 release"
    assert release.body == "release with changed name"


def test_create_release_asset(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    asset = release.create_asset(open("requirements.txt", "rb"))
    assert asset.name == "requirements.txt"
    assert asset.download_count == 0


def test_create_release_asset_with_name(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    asset = release.create_asset(open("requirements.txt", "rb"), "something else.txt")
    assert asset.name == "something else.txt"
    assert asset.download_count == 0


def test_get_release_assets(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    assert len(release.assets) > 0
    assert release.assets[0].name == "requirements.txt"


def test_download_release_asset(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    asset = release.assets[0]
    data = asset.download()
    assert data == open("requirements.txt", "rb").read()


def test_delete_release_asset(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    asset = release.assets[0]
    asset.delete()
    assert len(release.assets) == (len(repo.get_latest_release().assets) + 1)


def test_delete_release(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    release = repo.get_latest_release()
    old_releases = repo.get_releases()
    release.delete()
    assert len(repo.get_releases()) == len(old_releases) - 1


def test_create_commit_status(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    commit = repo.get_commits()[0]
    status = repo.create_commit_status(
        commit,
        state=CommitStatusState.ERROR,
        context="This is a test status",
    )
    assert status.status == CommitStatusState.ERROR
    assert status.context == "This is a test status"
    assert status.description == ""


def test_get_commit_statuses(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    commit = repo.get_commits()[0]
    statuses = commit.get_statuses()
    assert len(statuses) > 0
    assert statuses[0].context == "This is a test status"


def test_get_commit_combined_status(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    commit = repo.get_commits()[0]
    status = commit.get_status()
    assert status is not None
    assert status.state == CommitStatusState.FAILURE


def test_get_commit_status_from(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    commit = repo.get_commits()[0]
    statuses = repo.get_commit_statuses(commit, state=CommitStatusState.SUCCESS)
    assert len(statuses) == 0
    statuses = repo.get_commit_statuses(commit, state=CommitStatusState.ERROR)
    assert len(statuses) == 1


def test_get_repo_archive(instance):
    # This requires a repo with actual files in it, so we use the test repo
    repo = Repository.request(instance, test_org, test_repo)
    branch = repo.get_branches()[0]
    archive = repo.get_archive(branch)
    assert archive is not None


def test_team_get_org(instance):
    org = Organization.request(instance, test_org)
    team = org.get_team(test_team)
    assert org.username == team.organization.name


def test_delete_repo_userowned(instance):
    user = User.request(instance, test_user)
    repo = Repository.request(instance, user.username, test_repo)
    repo.delete()
    with pytest.raises(NotFoundException):
        Repository.request(instance, test_user, test_repo)


def test_secondary_email(instance):
    SECONDARYMAIL = "secondarytest@test.org"  # set up with real email
    sec_user = instance.get_user_by_email(SECONDARYMAIL)
    assert SECONDARYMAIL in sec_user.emails
    assert sec_user.username == "test"


def test_delete_repo_orgowned(instance):
    org = Organization.request(instance, test_org)
    repo = Repository.request(instance, org.username, test_repo)
    repo.delete()
    with pytest.raises(NotFoundException):
        Repository.request(instance, test_user, test_repo)


def test_change_repo_ownership_org(instance):
    old_org = Organization.request(instance, test_org)
    user = User.request(instance, test_user)
    new_org = instance.create_org(
        user, test_org + "_repomove", "Org for testing moving repositories"
    )
    new_team = instance.create_team(new_org, test_team + "_repomove", "descr")
    repo_name = test_repo + "_repomove"
    repo = instance.create_repo(old_org, repo_name, "descr")
    repo.transfer_ownership(new_org, set([new_team]))
    assert repo_name not in [repo.name for repo in old_org.get_repositories()]
    assert repo_name in [repo.name for repo in new_org.get_repositories()]


def test_change_repo_ownership_user(instance):
    old_org = Organization.request(instance, test_org)
    user = User.request(instance, test_user)
    repo_name = test_repo + "_repomove"
    repo = instance.create_repo(old_org, repo_name, "descr")
    repo.transfer_ownership(user)
    assert repo_name not in [repo.name for repo in old_org.get_repositories()]
    assert repo_name in [repo.name for repo in user.get_repositories()]
    for repo in user.get_repositories():
        repo.transfer_ownership(old_org)
        assert repo_name in [repo.name for repo in old_org.get_repositories()]
        assert repo_name not in [repo.name for repo in user.get_repositories()]


def test_delete_team(instance):
    org = Organization.request(instance, test_org)
    team = org.get_team(test_team)
    team.delete()
    with pytest.raises(NotFoundException):
        team = org.get_team(test_team)


def test_delete_teams(instance):
    org = Organization.request(instance, test_org)
    repos = org.get_repositories()
    for repo in repos:
        repo.delete()
    repos = org.get_repositories()
    assert len(repos) == 0


def test_delete_org(instance):
    org = Organization.request(instance, test_org)
    org.delete()
    with pytest.raises(NotFoundException):
        Organization.request(instance, test_org)


def test_delete_user(instance):
    user_name = test_user + "delte_test"
    email = user_name + "@example.org"
    user = instance.create_user(user_name, email, "abcdefg1.23AB", send_notify=False)
    assert user.username == user_name
    user.delete()
    with pytest.raises(NotFoundException):
        User.request(instance, user_name)
